@@ -62,19 +62,13 @@ STDCALL UCHAR WRAP_EXPORT(READ_PORT_UCHAR)
 STDCALL void WRAP_EXPORT(WRITE_PORT_BUFFER_USHORT)
 	(ULONG_PTR port, USHORT *buf, ULONG count)
 {
-	ULONG i;
-
-	for (i = 0 ; i < count ; i++)
-		outw(buf[i], port);
+	outsw(port, buf, count);
 }
 
 STDCALL void WRAP_EXPORT(READ_PORT_BUFFER_USHORT)
 	(ULONG_PTR port, USHORT *buf, ULONG count)
 {
-	ULONG i;
-
-	for (i = 0 ; i < count; i++)
-		buf[i] = inw(port);
+	insw(port, buf, count);
 }
 
 STDCALL void WRAP_EXPORT(WRITE_REGISTER_ULONG)
@@ -112,17 +106,17 @@ _FASTCALL KIRQL WRAP_EXPORT(KfRaiseIrql)
 {
 	KIRQL irql;
 
-	TRACEENTER4("irql = %d", newirql);
+	TRACEENTER5("irql = %d", newirql);
 	irql = raise_irql(newirql);
-	TRACEEXIT4(return irql);
+	TRACEEXIT5(return irql);
 }
 	
 _FASTCALL void WRAP_EXPORT(KfLowerIrql)
 	(FASTCALL_DECL_1(KIRQL oldirql))
 {
-	TRACEENTER4("irql = %d", oldirql);
+	TRACEENTER5("irql = %d", oldirql);
 	lower_irql(oldirql);
-	TRACEEXIT4(return);
+	TRACEEXIT5(return);
 }
 
 _FASTCALL KIRQL WRAP_EXPORT(KfAcquireSpinLock)
@@ -130,44 +124,41 @@ _FASTCALL KIRQL WRAP_EXPORT(KfAcquireSpinLock)
 {
 	KIRQL oldirql;
 
-	TRACEENTER4("lock = %p", lock);
+	TRACEENTER5("lock = %p", lock);
 	oldirql = kspin_lock_irql(lock, DISPATCH_LEVEL);
-	TRACEEXIT4(return oldirql);
+	TRACEEXIT5(return oldirql);
 }
 
 _FASTCALL void WRAP_EXPORT(KfReleaseSpinLock)
 	(FASTCALL_DECL_2(KSPIN_LOCK *lock, KIRQL newirql))
 {
-	TRACEENTER4("lock = %p, irql = %d", lock, newirql);
+	TRACEENTER5("lock = %p, irql = %d", lock, newirql);
 	kspin_unlock_irql(lock, newirql);
-	TRACEEXIT4(return);
+	TRACEEXIT5(return);
 }
 
 _FASTCALL void WRAP_EXPORT(KefAcquireSpinLockAtDpcLevel)
 	(FASTCALL_DECL_1(KSPIN_LOCK *lock))
 {
-	KIRQL irql;
-
-	TRACEENTER4("lock = %p", lock);
-#if DEBUG_IRQL
-	irql = current_irql();
-	if (irql != DISPATCH_LEVEL)
-		ERROR("irql %d != DISPATCH_LEVEL", irql);
+	TRACEENTER5("lock = %p", lock);
+#ifdef DEBUG_IRQL
+	if (current_irql() != DISPATCH_LEVEL)
+		ERROR("irql != DISPATCH_LEVEL");
 #endif
 	kspin_lock(lock);
-	TRACEEXIT4(return);
+	TRACEEXIT5(return);
 }
 
 _FASTCALL void WRAP_EXPORT(KefReleaseSpinLockFromDpcLevel)
 	(FASTCALL_DECL_1(KSPIN_LOCK *lock))
 {
-	TRACEENTER4("lock = %p", lock);
+	TRACEENTER5("lock = %p", lock);
 #ifdef DEBUG_IRQL
 	if (current_irql() != DISPATCH_LEVEL)
 		ERROR("irql != DISPATCH_LEVEL");
 #endif
 	kspin_unlock(lock);
-	TRACEEXIT4(return);
+	TRACEEXIT5(return);
 }
 
 #include "hal_exports.h"
