@@ -128,11 +128,6 @@ enum ndis_power {
 	NDIS_POWER_MIN,
 };
 
-enum ndis_power_profile {
-	NdisPowerProfileBattery,
-	NdisPowerProfileAcOnLine
-};
-
 struct ndis_auth_req {
 	ULONG length;
 	mac_address bssid;
@@ -160,6 +155,8 @@ int set_infra_mode(struct wrapper_dev *wd,
 int get_ap_address(struct wrapper_dev *wd, mac_address mac);
 int set_auth_mode(struct wrapper_dev *wd, int auth_mode);
 int set_encr_mode(struct wrapper_dev *wd, int encr_mode);
+int get_auth_mode(struct wrapper_dev *wd);
+int get_encr_mode(struct wrapper_dev *wd);
 int set_privacy_filter(struct wrapper_dev *wd, int flags);
 int set_scan(struct wrapper_dev *wd);
 
@@ -168,6 +165,7 @@ int set_scan(struct wrapper_dev *wd);
 #define PRIV_RESET	 		SIOCIWFIRSTPRIV+16
 #define PRIV_POWER_PROFILE	 	SIOCIWFIRSTPRIV+17
 #define PRIV_NETWORK_TYPE	 	SIOCIWFIRSTPRIV+18
+#define PRIV_USB_RESET	 		SIOCIWFIRSTPRIV+19
 
 /* these have to match what is in wpa_supplicant */
 typedef enum { WPA_ALG_NONE, WPA_ALG_WEP, WPA_ALG_TKIP, WPA_ALG_CCMP } wpa_alg;
