@@ -13,39 +13,27 @@
  *
  */
 
-#include <linux/init.h>
-#include <linux/module.h>
-#include <linux/kernel.h>
-#include <linux/vmalloc.h>
-#include <linux/kmod.h>
-
-#include <linux/types.h>
-#include <linux/fs.h>
-#include <linux/errno.h>
-#include <linux/miscdevice.h>
-#include <linux/pci.h>
-
-#include <linux/netdevice.h>
-#include <linux/etherdevice.h>
-#include <linux/ethtool.h>
-#include <linux/if_arp.h>
-#include <net/iw_handler.h>
-#include <linux/rtnetlink.h>
-#include <asm/pgtable.h>
-#include <asm/uaccess.h>
-
 #include "ndis.h"
 #include "loader.h"
 #include "wrapper.h"
 
-static KSPIN_LOCK loader_lock;
+#include <linux/module.h>
+#include <linux/kmod.h>
+#include <linux/miscdevice.h>
+#include <asm/uaccess.h>
+
+KSPIN_LOCK loader_lock;
 static struct ndis_device *ndis_devices;
 static unsigned int num_ndis_devices;
-struct list_head ndis_drivers;
+struct nt_list ndis_drivers;
 static struct pci_device_id *ndiswrapper_pci_devices;
-static struct usb_device_id *ndiswrapper_usb_devices;
 static struct pci_driver ndiswrapper_pci_driver;
+#if defined(CONFIG_USB)
+static struct usb_device_id *ndiswrapper_usb_devices;
 static struct usb_driver ndiswrapper_usb_driver;
+#endif
+
+extern int debug;
 
 /* load driver for given device, if not already loaded */
 static struct ndis_driver *ndiswrapper_load_driver(struct ndis_device *device)
@@ -57,8 +45,9 @@ static struct ndis_driver *ndiswrapper_load_driver(struct ndis_device *device)
 	TRACEENTER1("device: %04X:%04X:%04X:%04X", device->vendor,
 		    device->device, device->subvendor, device->subdevice);
 	found = 0;
+	ndis_driver = NULL;
 	irql = kspin_lock_irql(&loader_lock, DISPATCH_LEVEL);
-	list_for_each_entry(ndis_driver, &ndis_drivers, list) {
+	nt_list_for_each_entry(ndis_driver, &ndis_drivers, list) {
 		if (strcmp(ndis_driver->name, device->driver_name) == 0) {
 			DBGTRACE1("driver %s already loaded",
 				  ndis_driver->name);
@@ -77,7 +66,7 @@ static struct ndis_driver *ndiswrapper_load_driver(struct ndis_device *device)
 #else
 				"0",
 #endif
-				NDISWRAPPER_VERSION, device->driver_name,
+				UTILS_VERSION, device->driver_name,
 				device->conf_file_name, NULL};
 		char *env[] = {NULL};
 
@@ -101,7 +90,7 @@ static struct ndis_driver *ndiswrapper_load_driver(struct ndis_device *device)
 #endif
 		found = 0;
 		irql = kspin_lock_irql(&loader_lock, DISPATCH_LEVEL);
-		list_for_each_entry(ndis_driver, &ndis_drivers, list) {
+		nt_list_for_each_entry(ndis_driver, &ndis_drivers, list) {
 			if (strcmp(ndis_driver->name,
 				   device->driver_name) == 0) {
 				found = 1;
@@ -135,7 +124,6 @@ static int ndiswrapper_add_pci_device(struct pci_dev *pdev,
 	struct ndis_driver *driver;
 	struct wrapper_dev *wd;
 	struct net_device *dev;
-	struct miniport_char *miniport;
 	struct device_object *pdo;
 	struct driver_object *drv_obj;
 
@@ -145,41 +133,38 @@ static int ndiswrapper_add_pci_device(struct pci_dev *pdev,
 		  pdev->subsystem_vendor, pdev->subsystem_device);
 
 	device = &ndis_devices[ent->driver_data];
-
 	driver = ndiswrapper_load_driver(device);
-	if (!driver) {
-		res = -ENODEV;
-		goto out_nodev;
-	}
+	if (!driver)
+		return -ENODEV;
 
-	dev = ndis_init_netdev(&wd, device, driver);
+	dev = init_netdev(&wd, device, driver);
 	if (!dev) {
 		ERROR("couldn't initialize network device");
-		res = -ENOMEM;
-		goto out_nodev;
+		return -ENOMEM;
 	}
-	device->wd = wd;
 	DBGTRACE1("");
 	/* first create pdo */
 	drv_obj = find_bus_driver("PCI");
 	if (!drv_obj)
-		goto out_start;
+		goto err_bus_driver;
 	wd->dev.dev_type = NDIS_PCI_BUS;
 	wd->dev.pci = pdev;
 	DBGTRACE1("");
-	pdo = alloc_pdo(drv_obj, &wd->dev);
+	pdo = alloc_pdo(drv_obj);
 	if (!pdo) {
 		res = -ENODEV;
-		goto out_start;
+		goto err_bus_driver;
 	}
-	pdo->wd = wd;
+	pdo->reserved = wd;
+	wd->nmb->pdo = pdo;
+	DBGTRACE1("driver: %p", pdo->drv_obj);
 
-	DBGTRACE1("");
 	/* this creates (empty) fdo */
 	res = driver->drv_obj->drv_ext->add_device_func(driver->drv_obj,
 							pdo);
 	if (res != STATUS_SUCCESS)
-		goto out_start;
+		goto err_bus_driver;
+	DBGTRACE1("fdo: %p", wd->nmb->fdo);
 
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 	SET_NETDEV_DEV(dev, &pdev->dev);
@@ -190,20 +175,18 @@ static int ndiswrapper_add_pci_device(struct pci_dev *pdev,
 	res = pci_enable_device(pdev);
 	if (res) {
 		ERROR("couldn't enable PCI device: %08x", res);
-		goto out_enable;
+		goto err_bus_driver;
 	}
 
 	res = pci_request_regions(pdev, DRIVER_NAME);
 	if (res) {
 		ERROR("couldn't request PCI regions: %08x", res);
-		goto out_regions;
+		goto err_enable;
 	}
 
-	pci_set_power_state(pdev, 0);
-#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9)
-	pci_restore_state(pdev, NULL);
-#endif
+	res = pci_set_power_state(pdev, PCI_D0);
 
+#ifdef CONFIG_X86_64
 	/* 64-bit broadcom driver doesn't work if DMA is allocated
 	 * from over 1GB */
 	if (strcmp(device->driver_name, "netbc564") == 0) {
@@ -212,56 +195,24 @@ static int ndiswrapper_add_pci_device(struct pci_dev *pdev,
 			WARNING("DMA mask couldn't be set; this driver "
 				"may not work with more than 1GB RAM");
 	}
-	DBGTRACE1("%s", "calling ndis init routine");
-	if ((res = miniport_init(wd))) {
-		ERROR("Windows driver couldn't initialize the device (%08X)",
-			res);
+#endif
+	if (ndiswrapper_start_device(wd)) {
+		ERROR("couldn't start device");
 		res = -EINVAL;
-		goto out_start;
-	}
-
-	/* do we need to power up the card explicitly? */
-	miniport_set_int(wd, OID_PNP_SET_POWER, NdisDeviceStateD0);
-	miniport = &wd->driver->miniport;
-	/* According NDIS, pnp_event_notify should be called whenever power
-	 * is set to D0
-	 * Only NDIS 5.1 drivers are required to supply this function; some
-	 * drivers don't seem to support it (at least Orinoco)
-	 */
-	/*
-	if (miniport->pnp_event_notify) {
-		DBGTRACE3("%s", "calling pnp_event_notify");
-		miniport->pnp_event_notify(nmb, NDIS_PNP_PROFILE_CHANGED,
-					 &profile_inf, sizeof(profile_inf));
-	}
-	*/
-
-	/* IPW2200 devices turn off radio if reset is called */
-	if (pdev->vendor != 0x8086)
-		miniport_reset(wd);
-
-	/* Wait a little to let card power up otherwise ifup might fail after
-	   boot */
-	set_current_state(TASK_INTERRUPTIBLE);
-	schedule_timeout(HZ/2);
-
-	if (setup_device(wd->net_dev)) {
-		ERROR("couldn't setup network device");
-		res = -EINVAL;
-		goto out_setup;
+		goto err_regions;
 	}
 	atomic_inc(&driver->users);
+	device->wd = wd;
+	wd->ndis_device = device;
 	TRACEEXIT1(return 0);
 
-out_setup:
-	miniport_halt(wd);
-out_start:
+err_regions:
 	pci_release_regions(pdev);
-out_regions:
+err_enable:
 	pci_disable_device(pdev);
-out_enable:
+err_bus_driver:
+	pci_set_drvdata(pdev, NULL);
 	free_netdev(dev);
-out_nodev:
 	TRACEEXIT1(return res);
 }
 
@@ -280,13 +231,7 @@ static void __devexit ndiswrapper_remove_pci_device(struct pci_dev *pdev)
 
 	if (!wd)
 		TRACEEXIT1(return);
-
-	atomic_dec(&wd->driver->users);
-	ndiswrapper_remove_device(wd);
-
-	pci_release_regions(pdev);
-	pci_disable_device(pdev);
-	pci_set_drvdata(pdev, NULL);
+	ndiswrapper_stop_device(wd);
 }
 
 #ifdef CONFIG_USB
@@ -304,93 +249,80 @@ static void *ndiswrapper_add_usb_device(struct usb_device *udev,
 	struct ndis_driver *driver;
 	struct wrapper_dev *wd;
 	struct net_device *dev;
-	struct miniport_char *miniport;
 	struct device_object *pdo;
 	struct driver_object *drv_obj;
-//	unsigned long profile_inf = NDIS_POWER_PROFILE_AC;
 
-	TRACEENTER1("vendor: %04x, product: %04x",
-		    usb_id->idVendor, usb_id->idProduct);
+	TRACEENTER1("vendor: %04x, product: %04x, id: %p",
+		    usb_id->idVendor, usb_id->idProduct, usb_id);
 
 	device = &ndis_devices[usb_id->driver_info];
-	driver = ndiswrapper_load_driver(device);
-	if (!driver) {
-		res = -ENODEV;
-		goto out_nodev;
-	}
-	dev = ndis_init_netdev(&wd, device, driver);
-	if (!dev) {
-		ERROR("couldn't initialize network device");
-		res = -ENOMEM;
-		goto out_nodev;
+	/* RNDIS devices have two interfaces, so prevent from
+	 * initializing the device again, if it has already been
+	 * initialized */
+	if (device->wd) {
+		DBGTRACE1("device is already loaded");
+		TRACEEXIT1(return 0);
 	}
 
-	device->wd = wd;
+	driver = ndiswrapper_load_driver(device);
+	if (!driver)
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
+		return -ENODEV;
+#else
+		return NULL;
+#endif
+	dev = init_netdev(&wd, device, driver);
+	if (!dev) {
+		ERROR("couldn't initialize network device");
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
+		return -ENOMEM;
+#else
+		return NULL;
+#endif
+	}
 
 	/* first create pdo */
 	drv_obj = find_bus_driver("USB");
 	if (!drv_obj)
-		goto out_start;
+		goto err_net_dev;
 	wd->dev.dev_type = NDIS_USB_BUS;
 	DBGTRACE1("");
-	pdo = alloc_pdo(drv_obj, &wd->dev);
+	pdo = alloc_pdo(drv_obj);
 	if (!pdo) {
 		res = -ENODEV;
-		goto out_start;
+		goto err_net_dev;
 	}
-	pdo->wd = wd;
+	pdo->reserved = wd;
+	wd->nmb->pdo = pdo;
 
 	DBGTRACE1("");
 	/* this creates (empty) fdo */
 	res = driver->drv_obj->drv_ext->add_device_func(driver->drv_obj,
 							pdo);
 	if (res != STATUS_SUCCESS)
-		goto out_start;
+		goto err_pdo;
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 	SET_NETDEV_DEV(dev, &intf->dev);
 
-	wd->dev.usb = interface_to_usbdev(intf);
-	wd->intf    = intf;
+	wd->dev.usb.udev = interface_to_usbdev(intf);
 	usb_set_intfdata(intf, wd);
+	wd->dev.usb.intf = intf;
 #else
-	wd->dev.usb = udev;
+	wd->dev.usb.udev = udev;
+	wd->dev.usb.intf = usb_ifnum_to_if(udev, ifnum);
 #endif
-	pdo->device.usb = wd->dev.usb;
 
 	TRACEENTER1("calling ndis init routine");
-	if ((res = miniport_init(wd))) {
-		ERROR("Windows driver couldn't initialize the device (%08X)",
-			res);
+
+	if (ndiswrapper_start_device(wd)) {
+		ERROR("couldn't start device");
 		res = -EINVAL;
-		goto out_start;
-	}
-
-	/* do we need to power up the card explicitly? */
-	miniport_set_int(wd, OID_PNP_SET_POWER, NdisDeviceStateD0);
-	miniport = &wd->driver->miniport;
-	/*
-	if (miniport->pnp_event_notify) {
-		DBGTRACE3("%s", "calling pnp_event_notify");
-		miniport->pnp_event_notify(wd->adapter_ctx,
-					   NDIS_PNP_PROFILE_CHANGED,
-					   &profile_inf, sizeof(profile_inf));
-		DBGTRACE3("%s", "done");
-	}
-	*/
-
-	miniport_reset(wd);
-	/* wait here seems crucial; without this delay, at least
-	 * prism54 driver crashes (why?) */
-	set_current_state(TASK_INTERRUPTIBLE);
-	schedule_timeout(3*HZ);
-
-	if (setup_device(wd->net_dev)) {
-		ERROR("couldn't setup network device");
-		res = -EINVAL;
-		goto out_setup;
+		goto err_add_dev;
 	}
 
 	atomic_inc(&driver->users);
+	device->wd = wd;
+	wd->ndis_device = device;
 
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 	TRACEEXIT1(return 0);
@@ -398,11 +330,13 @@ static void *ndiswrapper_add_usb_device(struct usb_device *udev,
 	TRACEEXIT1(return wd);
 #endif
 
-out_setup:
-	miniport_halt(wd);
-out_start:
+err_add_dev:
+	DeleteDevice(pdo);
+err_pdo:
+	IoDeleteDevice(pdo);
+err_net_dev:
 	free_netdev(dev);
-out_nodev:
+	device->wd = NULL;
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 	TRACEEXIT1(return res);
 #else
@@ -420,12 +354,14 @@ ndiswrapper_remove_usb_device(struct usb_interface *intf)
 
 	TRACEENTER1("");
 	wd = (struct wrapper_dev *)usb_get_intfdata(intf);
-
 	if (!wd)
 		TRACEEXIT1(return);
+
+	if (!test_bit(HW_RMMOD, &wd->hw_status))
+		miniport_surprise_remove(wd);
+	wd->dev.usb.intf = NULL;
 	usb_set_intfdata(intf, NULL);
-	atomic_dec(&wd->driver->users);
-	ndiswrapper_remove_device(wd);
+	ndiswrapper_stop_device(wd);
 }
 #else
 static void
@@ -434,12 +370,15 @@ ndiswrapper_remove_usb_device(struct usb_device *udev, void *ptr)
 	struct wrapper_dev *wd = (struct wrapper_dev *)ptr;
 
 	TRACEENTER1("");
-
-	if (!wd || !wd->dev.usb)
+	if (!wd || !wd->dev.usb.intf)
 		TRACEEXIT1(return);
-	wd->dev.usb = NULL;
+	if (!test_bit(HW_RMMOD, &wd->hw_status))
+		miniport_surprise_remove(wd);
+	wd->dev.usb.intf = NULL;
 	atomic_dec(&wd->driver->users);
-	ndiswrapper_remove_device(wd);
+	if (wd->ndis_device)
+		wd->ndis_device->wd = NULL;
+	ndiswrapper_stop_device(wd);
 }
 #endif
 #endif /* CONFIG_USB */
@@ -517,7 +456,7 @@ static int load_sys_files(struct ndis_driver *driver,
 	}
 
 	if (load_pe_images(driver->pe_images, driver->num_pe_images)) {
-		ERROR("unable to prepare driver '%s'", load_driver->name);
+		ERROR("couldn't prepare driver '%s'", load_driver->name);
 		err = -EINVAL;
 	}
 
@@ -586,31 +525,30 @@ static int load_bin_files(struct ndis_driver *driver,
 static int load_settings(struct ndis_driver *ndis_driver,
 			 struct load_driver *load_driver)
 {
-	int i, found, nr_settings;
+	int i, nr_settings;
 	struct ndis_device *ndis_device;
 	KIRQL irql;
 
 	TRACEENTER1("");
 
-	found = 0;
+	ndis_device = NULL;
 	irql = kspin_lock_irql(&loader_lock, DISPATCH_LEVEL);
 	for (i = 0; i < num_ndis_devices; i++) {
 		if (strcmp(ndis_devices[i].conf_file_name,
 			   load_driver->conf_file_name) == 0) {
-			found = 1;
+			ndis_device = &ndis_devices[i];
 			break;
 		}
 	}
 	kspin_unlock_irql(&loader_lock, irql);
 
-	if (!found) {
+	if (!ndis_device) {
 		ERROR("conf file %s not found",
 		      ndis_devices[i].conf_file_name);
 		TRACEEXIT1(return -EINVAL);
 	}
 
 	nr_settings = 0;
-	ndis_device = &ndis_devices[i];
 	for (i = 0; i < load_driver->nr_settings; i++) {
 		struct load_device_setting *load_setting =
 			&load_driver->settings[i];
@@ -633,7 +571,7 @@ static int load_settings(struct ndis_driver *ndis_driver,
 			memcpy(ndis_driver->version, setting->value,
 			       sizeof(ndis_driver->version));
 		irql = kspin_lock_irql(&loader_lock, DISPATCH_LEVEL);
-		list_add(&setting->list, &ndis_device->settings);
+		InsertTailList(&ndis_device->settings, &setting->list);
 		kspin_unlock_irql(&loader_lock, irql);
 		nr_settings++;
 	}
@@ -647,6 +585,7 @@ static int load_settings(struct ndis_driver *ndis_driver,
 /* this function is called while holding load_lock spinlock */
 static void unload_ndis_device(struct ndis_device *device)
 {
+	struct nt_list *cur;
 	TRACEENTER1("unloading device %p (%04X:%04X:%04X:%04X), driver %s",
 		    device, device->vendor, device->device, device->subvendor,
 		    device->subdevice, device->driver_name);
@@ -654,16 +593,14 @@ static void unload_ndis_device(struct ndis_device *device)
 	DBGTRACE3("%p", device->ndis_driver);
 	if (!device->ndis_driver)
 		TRACEEXIT1(return);
-	while (!list_empty(&device->settings)) {
+	while ((cur = RemoveHeadList(&device->settings))) {
 		struct device_setting *setting;
 		struct ndis_config_param *param;
 
-		setting = list_entry(device->settings.next,
-				     struct device_setting, list);
+		setting = container_of(cur, struct device_setting, list);
 		param = &setting->config_param;
 		if (param->type == NDIS_CONFIG_PARAM_STRING)
 			RtlFreeUnicodeString(&param->data.ustring);
-		list_del(&setting->list);
 		kfree(setting);
 	}
 	TRACEEXIT1(return);
@@ -679,16 +616,21 @@ static void unload_ndis_driver(struct ndis_driver *driver)
 	DBGTRACE1("freeing %d images", driver->num_pe_images);
 	drv_obj = driver->drv_obj;
 	for (i = 0; i < driver->num_pe_images; i++)
-		if (driver->pe_images[i].image)
+		if (driver->pe_images[i].image) {
+			DBGTRACE1("freeing image at %p",
+				  driver->pe_images[i].image);
 			vfree(driver->pe_images[i].image);
+		}
 
 	DBGTRACE1("freeing %d bin files", driver->num_bin_files);
-	for (i = 0; i < driver->num_bin_files; i++)
+	for (i = 0; i < driver->num_bin_files; i++) {
+		DBGTRACE1("freeing image at %p", driver->bin_files[i].data);
 		vfree(driver->bin_files[i].data);
+	}
 	if (driver->bin_files)
 		kfree(driver->bin_files);
 
-	IoDetachDevice(drv_obj->dev_obj->next);
+//	IoDetachDevice(drv_obj->dev_obj);
 	RtlFreeUnicodeString(&drv_obj->driver_name);
 	/* this frees driver */
 	free_custom_ext(drv_obj->drv_ext);
@@ -700,7 +642,8 @@ static void unload_ndis_driver(struct ndis_driver *driver)
 /* call the entry point of the driver */
 static int start_driver(struct ndis_driver *driver)
 {
-	int i, ret, res;
+	int i;
+	NTSTATUS ret, res;
 	struct driver_object *drv_obj;
 	UINT (*entry)(struct driver_object *obj,
 		      struct unicode_string *path) STDCALL;
@@ -719,8 +662,8 @@ static int start_driver(struct ndis_driver *driver)
 			ret |= res;
 			DBGTRACE1("entry returns %08X", res);
 			DBGTRACE1("driver version: %d.%d",
-				  driver->miniport.majorVersion,
-				  driver->miniport.minorVersion);
+				  driver->miniport.major_version,
+				  driver->miniport.minor_version);
 			break;
 		}
 
@@ -743,19 +686,19 @@ static int start_driver(struct ndis_driver *driver)
  */
 static int add_driver(struct ndis_driver *driver)
 {
-	struct ndis_driver *tmp;
 	KIRQL irql;
+	struct ndis_driver *tmp;
 
 	TRACEENTER1("");
 	irql = kspin_lock_irql(&loader_lock, DISPATCH_LEVEL);
-	list_for_each_entry(tmp, &ndis_drivers, list) {
+	nt_list_for_each_entry(tmp, &ndis_drivers, list) {
 		if (strcmp(tmp->name, driver->name) == 0) {
 			kspin_unlock_irql(&loader_lock, irql);
 			ERROR("cannot add duplicate driver");
 			TRACEEXIT1(return -EBUSY);
 		}
 	}
-	list_add(&driver->list, &ndis_drivers);
+	InsertTailList(&ndis_drivers, &driver->list);
 	kspin_unlock_irql(&loader_lock, irql);
 
 	TRACEEXIT1(return 0);
@@ -767,22 +710,24 @@ static int load_ndis_driver(struct load_driver *load_driver)
 	struct driver_object *drv_obj;
 	struct ansi_string ansi_reg;
 	struct ndis_driver *ndis_driver = NULL;
+	int i;
 
 	TRACEENTER1("");
-	ansi_reg.buf = "/tmp";
-	ansi_reg.buflen = ansi_reg.len = strlen(ansi_reg.buf);
 	drv_obj = kmalloc(sizeof(*drv_obj), GFP_KERNEL);
 	if (!drv_obj) {
 		ERROR("couldn't allocate memory");
 		TRACEEXIT1(return -ENOMEM);
 	}
 	DBGTRACE1("drv_obj: %p", drv_obj);
+	memset(drv_obj, 0, sizeof(*drv_obj));
 	drv_obj->drv_ext = kmalloc(sizeof(*(drv_obj->drv_ext)), GFP_KERNEL);
 	if (!drv_obj->drv_ext) {
 		ERROR("couldn't allocate memory");
 		kfree(drv_obj);
 		TRACEEXIT1(return -ENOMEM);
 	}
+	memset(drv_obj->drv_ext, 0, sizeof(*(drv_obj->drv_ext)));
+	drv_obj->drv_ext->add_device_func = AddDevice;
 	InitializeListHead(&drv_obj->drv_ext->custom_ext);
 	DBGTRACE1("");
 	if (IoAllocateDriverObjectExtension(drv_obj,
@@ -792,9 +737,16 @@ static int load_ndis_driver(struct load_driver *load_driver)
 	    STATUS_SUCCESS)
 		TRACEEXIT1(return NDIS_STATUS_RESOURCES);
 	DBGTRACE1("driver: %p", ndis_driver);
+	for (i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++)
+		drv_obj->major_func[i] = IopPassIrpDown;
+//	drv_obj->major_func[IRP_MJ_PNP] = fdoDispatchPnp;
+
 	memset(ndis_driver, 0, sizeof(*ndis_driver));
 	ndis_driver->bustype = -1;
 	ndis_driver->drv_obj = drv_obj;
+	ansi_reg.buf = "/tmp";
+	ansi_reg.length = strlen(ansi_reg.buf);
+	ansi_reg.max_length = ansi_reg.length + 1;
 	if (RtlAnsiStringToUnicodeString(&drv_obj->driver_name,
 					 &ansi_reg, 1) != STATUS_SUCCESS) {
 		ERROR("couldn't initialize registry path");
@@ -831,7 +783,9 @@ static int register_devices(struct load_devices *load_devices)
 
 	devices = NULL;
 	ndiswrapper_pci_devices = NULL;
+#if defined(CONFIG_USB)
 	ndiswrapper_usb_devices = NULL;
+#endif
 	ndis_devices = NULL;
 	devices = vmalloc(load_devices->count * sizeof(struct load_device));
 	if (!devices) {
@@ -867,6 +821,7 @@ static int register_devices(struct load_devices *load_devices)
 		       (num_pci + 1) * sizeof(struct pci_device_id));
 	}
 
+#if defined(CONFIG_USB)
 	if (num_usb > 0) {
 		ndiswrapper_usb_devices =
 			kmalloc((num_usb + 1) * sizeof(struct usb_device_id),
@@ -878,6 +833,7 @@ static int register_devices(struct load_devices *load_devices)
 		memset(ndiswrapper_usb_devices, 0,
 		       (num_usb + 1) * sizeof(struct usb_device_id));
 	}
+#endif
 
 	ndis_devices = vmalloc(num_ndis_devices * sizeof(*ndis_devices));
 	if (!ndis_devices) {
@@ -893,7 +849,7 @@ static int register_devices(struct load_devices *load_devices)
 
 		ndis_device = &ndis_devices[num_pci + num_usb];
 
-		INIT_LIST_HEAD(&ndis_device->settings);
+		InitializeListHead(&ndis_device->settings);
 		memcpy(&ndis_device->driver_name, device->driver_name,
 		       sizeof(ndis_device->driver_name));
 		memcpy(&ndis_device->conf_file_name, device->conf_file_name,
@@ -947,8 +903,7 @@ static int register_devices(struct load_devices *load_devices)
 				  device->vendor, device->device);
 #endif
 		} else {
-			ERROR("system doesn't support bus type %d",
-			      device->bustype);
+			ERROR("bus type %d not supported", device->bustype);
 		}
 	}
 
@@ -978,6 +933,10 @@ static int register_devices(struct load_devices *load_devices)
 		ndiswrapper_usb_driver.probe = ndiswrapper_add_usb_device;
 		ndiswrapper_usb_driver.disconnect =
 			ndiswrapper_remove_usb_device;
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
+		ndiswrapper_usb_driver.suspend = ndiswrapper_suspend_usb;
+		ndiswrapper_usb_driver.resume = ndiswrapper_resume_usb;
+#endif
 		res = usb_register(&ndiswrapper_usb_driver);
 		if (res < 0) {
 			ERROR("couldn't register ndiswrapper usb driver");
@@ -993,9 +952,11 @@ err:
 	if (ndis_devices)
 		vfree(ndis_devices);
 	ndis_devices = NULL;
+#if defined(CONFIG_USB)
 	if (ndiswrapper_usb_devices)
 		kfree(ndiswrapper_usb_devices);
 	ndiswrapper_usb_devices = NULL;
+#endif
 	if (ndiswrapper_pci_devices)
 		kfree(ndiswrapper_pci_devices);
 	ndiswrapper_pci_devices = NULL;
@@ -1072,7 +1033,7 @@ int loader_init(void)
 {
 	int err;
 
-	INIT_LIST_HEAD(&ndis_drivers);
+	InitializeListHead(&ndis_drivers);
 	kspin_lock_init(&loader_lock);
 	if ((err = misc_register(&wrapper_misc)) < 0 ) {
 		ERROR("couldn't register module (%d)", err);
@@ -1084,10 +1045,14 @@ int loader_init(void)
 void loader_exit(void)
 {
 	int i;
+	struct nt_list *cur;
 
 	TRACEENTER1("");
 	misc_deregister(&wrapper_misc);
 
+	for (i = 0; i < num_ndis_devices; i++)
+		if (ndis_devices[i].wd)
+			set_bit(HW_RMMOD, &ndis_devices[i].wd->hw_status);
 #ifdef CONFIG_USB
 	if (ndiswrapper_usb_devices) {
 		usb_deregister(&ndiswrapper_usb_driver);
@@ -1109,12 +1074,10 @@ void loader_exit(void)
 		ndis_devices = NULL;
 	}
 
-	while (!list_empty(&ndis_drivers)) {
+	while ((cur = RemoveHeadList(&ndis_drivers))) {
 		struct ndis_driver *driver;
 
-		driver = list_entry(ndis_drivers.next,
-				    struct ndis_driver, list);
-		list_del(&driver->list);
+		driver = container_of(cur, struct ndis_driver, list);
 		unload_ndis_driver(driver);
 	}
 	kspin_unlock(&loader_lock);
