@@ -1,3 +1,4 @@
+
 /*
  *  Copyright (C) 2003-2005 Pontus Fuchs, Giridhar Pemmasani
  *
@@ -23,143 +24,73 @@
 
 #include "ndis.h"
 
-static struct list_head wrap_allocs;
+static struct nt_list wrap_allocs;
 static KSPIN_LOCK wrap_allocs_lock;
-static struct nt_list wrapper_timer_list;
-static struct timer_list shared_data_timer;
-
-extern KSPIN_LOCK ntoskernel_lock;
-extern struct work_struct kdpc_work;
-
-struct kuser_shared_data kuser_shared_data;
-static void update_user_shared_data_proc(unsigned long data);
 
 int misc_funcs_init(void)
 {
-	INIT_LIST_HEAD(&wrap_allocs);
+	InitializeListHead(&wrap_allocs);
 	kspin_lock_init(&wrap_allocs_lock);
-	InitializeListHead(&wrapper_timer_list);
+	return 0;
+}
 
-	memset(&kuser_shared_data, 0, sizeof(kuser_shared_data));
-	init_timer(&shared_data_timer);
-	shared_data_timer.function = &update_user_shared_data_proc;
-#if defined(CONFIG_X86_64) && defined(INPROCOMM_AMD64)
-	*((ULONG64 *)SHARED_SYSTEM_TIME) = ticks_1601();
-	shared_data_timer.data = (unsigned long)0;
-	shared_data_timer.expires = jiffies + 10 * HZ / 1000;
-	add_timer(&shared_data_timer);
-#endif
+int misc_funcs_init_device(struct wrapper_dev *wd)
+{
 	return 0;
 }
 
 /* called when a handle is being removed */
 void misc_funcs_exit_device(struct wrapper_dev *wd)
 {
-	BOOLEAN canceled;
-	KIRQL irql;
-
-	/* cancel any timers left by bugyy windows driver Also free
-	 * the memory for timers */
-	while (1) {
-		struct nt_list *ent;
-		struct wrapper_timer *wrapper_timer;
-
-		irql = kspin_lock_irql(&wd->timer_lock, DISPATCH_LEVEL);
-		ent = RemoveHeadList(&wd->wrapper_timer_list);
-		kspin_unlock_irql(&wd->timer_lock, irql);
-		if (!ent)
-			break;
-		wrapper_timer = container_of(ent, struct wrapper_timer, list);
-		wrapper_cancel_timer(wrapper_timer, &canceled);
-		if (canceled == TRUE)
-			WARNING("Buggy Windows driver '%s' left a timer (%p)"
-				"running - fixed it", wd->driver->name,
-				wrapper_timer->ktimer);
-		wrap_kfree(wrapper_timer);
-	}
+	return;
 }
 
 /* called when module is being removed */
 void misc_funcs_exit(void)
 {
 	KIRQL irql;
+	struct nt_list *ent;
 
-	/* free kernel (Ke) timers */
-	while (1) {
-		struct nt_list *ent;
-		struct wrapper_timer *wrapper_timer;
-		BOOLEAN canceled;
-
-		irql = kspin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
-		ent = RemoveTailList(&wrapper_timer_list);
-		kspin_unlock_irql(&ntoskernel_lock, irql);
-		if (!ent)
-			break;
-		wrapper_timer = container_of(ent, struct wrapper_timer, list);
-		wrapper_cancel_timer(wrapper_timer, &canceled);
-		if (canceled == TRUE)
-			WARNING("Buggy Windows driver left a timer (%p)"
-				"running - fixed it", wrapper_timer);
-		wrap_kfree(wrapper_timer);
-	}
 	/* free all pointers on the allocated list */
 	irql = kspin_lock_irql(&wrap_allocs_lock, DISPATCH_LEVEL);
-	while (!list_empty(&wrap_allocs)) {
+	while ((ent = RemoveHeadList(&wrap_allocs))) {
 		struct wrap_alloc *alloc;
-
-		alloc = list_entry(wrap_allocs.next, struct wrap_alloc, list);
-		list_del(&alloc->list);
+		alloc = container_of(ent, struct wrap_alloc, list);
 		kfree(alloc->ptr);
 		kfree(alloc);
 	}
 	kspin_unlock_irql(&wrap_allocs_lock, irql);
-
-#if defined(CONFIG_X86_64) && defined(INPROCOMM_AMD64)
-	del_timer_sync(&shared_data_timer);
-#endif
-
 	TRACEEXIT4(return);
 }
 
-static void update_user_shared_data_proc(unsigned long data)
-{
-#if defined(CONFIG_X86_64) && defined(INPROCOMM_AMD64)
-	/* timer is scheduled every 10ms and the system timer is in
-	 * 100ns */
-	*((ULONG64 *)SHARED_SYSTEM_TIME) = ticks_1601();
-	*((ULONG64 *)SHARED_INTERRUPT_TIME) = jiffies * TICKSPERSEC / HZ;
-	*((ULONG64 *)SHARED_TICK_COUNT) = jiffies;
-#endif
-	shared_data_timer.expires = jiffies + 10 * HZ / 1000;
-	add_timer(&shared_data_timer);
-}
-
 /* allocate memory with given flags and add it to list of allocated pointers;
  * if a driver doesn't free this memory for any reason (buggy driver or we
  * allocate space behind driver's back since we need more space than
  * corresponding Windows structure provides etc.), this gets freed
  * automatically during module unloading
  */
-void *wrap_kmalloc(size_t size, int flags)
+void *wrap_kmalloc(size_t size)
 {
 	struct wrap_alloc *alloc;
 	KIRQL irql;
+	unsigned int alloc_flags;
 
-	TRACEENTER4("size = %lu, flags = %d", (unsigned long)size, flags);
+	TRACEENTER4("size = %lu", (unsigned long)size);
 
-	if ((flags & GFP_ATOMIC) || irqs_disabled())
-		alloc = kmalloc(sizeof(*alloc), GFP_ATOMIC);
+	if (current_irql() < DISPATCH_LEVEL)
+		alloc_flags = GFP_KERNEL;
 	else
-		alloc = kmalloc(sizeof(*alloc), GFP_KERNEL);
+		alloc_flags = GFP_ATOMIC;
+	alloc = kmalloc(sizeof(*alloc), alloc_flags);
 	if (!alloc)
 		return NULL;
-	alloc->ptr = kmalloc(size, flags);
+	alloc->ptr = kmalloc(size, alloc_flags);
 	if (!alloc->ptr) {
 		kfree(alloc);
 		return NULL;
 	}
 	irql = kspin_lock_irql(&wrap_allocs_lock, DISPATCH_LEVEL);
-	list_add(&alloc->list, &wrap_allocs);
+	InsertTailList(&wrap_allocs, &alloc->list);
 	kspin_unlock_irql(&wrap_allocs_lock, irql);
 	DBGTRACE4("%p, %p", alloc, alloc->ptr);
 	TRACEEXIT4(return alloc->ptr);
@@ -168,17 +99,14 @@ void *wrap_kmalloc(size_t size, int flags)
 /* free pointer and remove from list of allocated pointers */
 void wrap_kfree(void *ptr)
 {
-	struct list_head *cur, *tmp;
+	struct wrap_alloc *alloc;
 	KIRQL irql;
 
 	TRACEENTER4("%p", ptr);
 	irql = kspin_lock_irql(&wrap_allocs_lock, DISPATCH_LEVEL);
-	list_for_each_safe(cur, tmp, &wrap_allocs) {
-		struct wrap_alloc *alloc;
-
-		alloc = list_entry(cur, struct wrap_alloc, list);
+	nt_list_for_each_entry(alloc, &wrap_allocs, list) {
 		if (alloc->ptr == ptr) {
-			list_del(&alloc->list);
+			RemoveEntryList(&alloc->list);
 			kfree(alloc->ptr);
 			kfree(alloc);
 			break;
@@ -188,187 +116,6 @@ void wrap_kfree(void *ptr)
 	TRACEEXIT4(return);
 }
 
-void wrapper_timer_handler(unsigned long data)
-{
-	struct ktimer *ktimer = (struct ktimer *)data;
-	struct wrapper_timer *wrapper_timer;
-	struct kdpc *kdpc;
-
-	TRACEENTER5("%p", ktimer);
-	wrapper_timer = ktimer->wrapper_timer;
-
-#ifdef DEBUG_TIMER
-	BUG_ON(wrapper_timer == NULL);
-	BUG_ON(wrapper_timer->wrapper_timer_magic != WRAPPER_TIMER_MAGIC);
-	BUG_ON(wrapper_timer->kdpc == NULL);
-#endif
-
-	kdpc = wrapper_timer->kdpc;
-	if (kdpc) {
-		kspin_lock(&ntoskernel_lock);
-		insert_kdpc_work(kdpc);
-		kspin_unlock(&ntoskernel_lock);
-	}
-	kspin_lock(&wrapper_timer->lock);
-	/* don't add the timer if aperiodic - see wrapper_cancel_timer;
-	 * protect access to kdpc, repeat, and active via spinlock */
-	if (wrapper_timer->repeat) {
-		wrapper_timer->timer.expires = jiffies + wrapper_timer->repeat;
-		add_timer(&wrapper_timer->timer);
-	} else
-		wrapper_timer->active = 0;
-	kspin_unlock(&wrapper_timer->lock);
-
-	KeSetEvent((struct kevent *)ktimer, 0, FALSE);
-
-	TRACEEXIT5(return);
-}
-
-/* we don't initialize ktimer event's signal here; that is caller's
- * responsibility */
-
-/* NDIS (miniport) timers set kdpc during timer initialization and Ke
- * timers set it during timer setup, so it is passed to both
- * wrapper_init_timer and wrapper_set_timer */
-void wrapper_init_timer(struct ktimer *ktimer, void *handle, struct kdpc *kdpc)
-{
-	struct wrapper_timer *wrapper_timer;
-	struct wrapper_dev *wd = (struct wrapper_dev *)handle;
-	KIRQL irql;
-
-	TRACEENTER5("%s", "");
-	/* we allocate memory for wrapper_timer behind driver's back
-	 * and there is no NDIS/DDK function where this memory can be
-	 * freed, so we use wrap_kmalloc so it gets freed when driver
-	 * is unloaded */
-	wrapper_timer = wrap_kmalloc(sizeof(*wrapper_timer), GFP_ATOMIC);
-	if (!wrapper_timer) {
-		ERROR("couldn't allocate memory for timer");
-		return;
-	}
-
-	memset(wrapper_timer, 0, sizeof(*wrapper_timer));
-	init_timer(&wrapper_timer->timer);
-	wrapper_timer->timer.data = (unsigned long)ktimer;
-	wrapper_timer->timer.function = &wrapper_timer_handler;
-#ifdef DEBUG_TIMER
-	wrapper_timer->wrapper_timer_magic = WRAPPER_TIMER_MAGIC;
-#endif
-	wrapper_timer->kdpc = kdpc;
-	wrapper_timer->ktimer = ktimer;
-	ktimer->wrapper_timer = wrapper_timer;
-	kspin_lock_init(&wrapper_timer->lock);
-	if (wd) {
-		irql = kspin_lock_irql(&wd->timer_lock, DISPATCH_LEVEL);
-		InsertHeadList(&wd->wrapper_timer_list,
-			       &wrapper_timer->list);
-		kspin_unlock_irql(&wd->timer_lock, irql);
-	} else {
-		irql = kspin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
-		InsertHeadList(&wrapper_timer_list, &wrapper_timer->list);
-		kspin_unlock_irql(&ntoskernel_lock, irql);
-	}
-
-	DBGTRACE4("added timer %p", wrapper_timer);
-	TRACEEXIT5(return);
-}
-
-/* 'expires' is relative to jiffies, so when setting timer, add
- * jiffies to it */
-int wrapper_set_timer(struct wrapper_timer *wrapper_timer,
-		      unsigned long expires, unsigned long repeat,
-		      struct kdpc *kdpc)
-{
-	KIRQL irql;
-	struct ktimer *ktimer = wrapper_timer->ktimer;
-
-	TRACEENTER5("%p", wrapper_timer);
-	if (!wrapper_timer) {
-		ERROR("invalid timer");
-		return FALSE;
-	}
-
-#ifdef DEBUG_TIMER
-	if (wrapper_timer->wrapper_timer_magic != WRAPPER_TIMER_MAGIC) {
-		WARNING("timer %p is not initialized (%lu)",
-			wrapper_timer, wrapper_timer->wrapper_timer_magic);
-		wrapper_timer->wrapper_timer_magic = WRAPPER_TIMER_MAGIC;
-	}
-#endif
-
-	/* timer handler also uses timer->repeat, active, and kdpc, so
-	 * protect in case of SMP */
-	KeClearEvent((struct kevent *)ktimer);
-	irql = kspin_lock_irql(&wrapper_timer->lock, DISPATCH_LEVEL);
-	if (kdpc)
-		wrapper_timer->kdpc = kdpc;
-	wrapper_timer->repeat = repeat;
-	if (wrapper_timer->active) {
-		DBGTRACE4("modifying timer %p to %lu, %lu",
-			  wrapper_timer, expires, repeat);
-		mod_timer(&wrapper_timer->timer, jiffies + expires);
-		kspin_unlock_irql(&wrapper_timer->lock, irql);
-		TRACEEXIT5(return TRUE);
-	} else {
-		DBGTRACE4("setting timer %p to %lu, %lu",
-			  wrapper_timer, expires, repeat);
-		wrapper_timer->timer.expires = jiffies + expires;
-		wrapper_timer->active = 1;
-		add_timer(&wrapper_timer->timer);
-		kspin_unlock_irql(&wrapper_timer->lock, irql);
-		TRACEEXIT5(return FALSE);
-	}
-}
-
-void wrapper_cancel_timer(struct wrapper_timer *wrapper_timer,
-			  BOOLEAN *canceled)
-{
-	KIRQL irql;
-	struct ktimer *ktimer = wrapper_timer->ktimer;
-
-	TRACEENTER4("timer = %p, canceled = %p", wrapper_timer, canceled);
-	if (!wrapper_timer) {
-		ERROR("%s", "invalid wrapper_timer");
-		return;
-	}
-
-#ifdef DEBUG_TIMER
-	DBGTRACE4("canceling timer %p", wrapper_timer);
-	BUG_ON(wrapper_timer->wrapper_timer_magic != WRAPPER_TIMER_MAGIC);
-#endif
-	/* timer handler also uses timer->repeat, so protect it in
-	 * case of SMP */
-
-	/* Oh, and del_timer_sync is not required ('canceled' argument
-	 * tells the driver if the timer was deleted or not) here; nor
-	 * is del_timer_sync correct, as this function may be called
-	 * at DISPATCH_LEVEL */
-	irql = raise_irql(DISPATCH_LEVEL);
-	kspin_lock(&ntoskernel_lock);
-	remove_kdpc_work(wrapper_timer->kdpc);
-	kspin_unlock(&ntoskernel_lock);
-	kspin_lock(&wrapper_timer->lock);
-	if (wrapper_timer->active) {
-		if (wrapper_timer->repeat) {
-			/* first mark as aperiodic, so timer function
-			 * doesn't call add_timer after del_timer
-			 * returned */
-			wrapper_timer->repeat = 0;
-			del_timer(&wrapper_timer->timer);
-			/* periodic timers always return TRUE */
-			*canceled = TRUE;
-		} else
-			*canceled = del_timer(&wrapper_timer->timer);
-		wrapper_timer->active = 0;
-	} else
-		*canceled = FALSE;
-	kspin_unlock(&wrapper_timer->lock);
-	lower_irql(irql);
-	if (*canceled == TRUE)
-		KeClearEvent((struct kevent *)ktimer);
-	TRACEEXIT5(return);
-}
-
 NOREGPARM INT WRAP_EXPORT(_win_sprintf)
 	(char *buf, const char *format, ...)
 {
@@ -377,13 +124,17 @@ NOREGPARM INT WRAP_EXPORT(_win_sprintf)
 	va_start(args, format);
 	res = vsprintf(buf, format, args);
 	va_end(args);
+	DBGTRACE2("buf: %p: %s", buf, buf);
 	return res;
 }
 
 NOREGPARM INT WRAP_EXPORT(_win_vsprintf)
 	(char *str, const char *format, va_list ap)
 {
-	return vsprintf(str, format, ap);
+	INT i;
+	i = vsprintf(str, format, ap);
+	DBGTRACE2("str: %p: %s", str, str);
+	TRACEEXIT2(return i);
 }
 
 NOREGPARM INT WRAP_EXPORT(_win_snprintf)
@@ -395,6 +146,7 @@ NOREGPARM INT WRAP_EXPORT(_win_snprintf)
 	va_start(args, format);
 	res = vsnprintf(buf, count, format, args);
 	va_end(args);
+	DBGTRACE2("buf: %p: %s", buf, buf);
 	return res;
 }
 
@@ -407,19 +159,26 @@ NOREGPARM INT WRAP_EXPORT(_win__snprintf)
 	va_start(args, format);
 	res = vsnprintf(buf, count, format, args);
 	va_end(args);
+	DBGTRACE2("buf: %p: %s", buf, buf);
 	return res;
 }
 
 NOREGPARM INT WRAP_EXPORT(_win_vsnprintf)
 	(char *str, SIZE_T size, const char *format, va_list ap)
 {
-	return vsnprintf(str, size, format, ap);
+	INT i;
+	i = vsnprintf(str, size, format, ap);
+	DBGTRACE2("str: %p: %s", str, str);
+	TRACEEXIT2(return i);
 }
 
 NOREGPARM INT WRAP_EXPORT(_win__vsnprintf)
 	(char *str, SIZE_T size, const char *format, va_list ap)
 {
-	return vsnprintf(str, size, format, ap);
+	INT i;
+	i = vsnprintf(str, size, format, ap);
+	DBGTRACE2("str: %p: %s", str, str);
+	TRACEEXIT2(return i);
 }
 
 NOREGPARM char *WRAP_EXPORT(_win_strncpy)
@@ -461,7 +220,7 @@ NOREGPARM char *WRAP_EXPORT(_win_strncat)
 NOREGPARM INT WRAP_EXPORT(_win_wcscmp)
 	(const wchar_t *s1, const wchar_t *s2)
 {
-	while (*s1 && *s2 && *s1 == *s2) {
+	while (*s1 != L'\0' && *s1 == *s2) {
 		s1++;
 		s2++;
 	}
@@ -471,7 +230,7 @@ NOREGPARM INT WRAP_EXPORT(_win_wcscmp)
 NOREGPARM INT WRAP_EXPORT(_win_wcsicmp)
 	(const wchar_t *s1, const wchar_t *s2)
 {
-	while (*s1 && *s2 && tolower((char)*s1) == tolower((char)*s2)) {
+	while (*s1 != L'\0' && tolower((char)*s1) == tolower((char)*s2)) {
 		s1++;
 		s2++;
 	}
@@ -482,7 +241,7 @@ NOREGPARM SIZE_T WRAP_EXPORT(_win_wcslen)
 	(const wchar_t *s)
 {
 	SIZE_T i = 0;
-	while (s[i])
+	while (s[i] != L'\0')
 		i++;
 	return i;
 }
@@ -491,7 +250,7 @@ NOREGPARM wchar_t *WRAP_EXPORT(_win_wcsncpy)
 	(wchar_t *dest, const wchar_t *src, SIZE_T n)
 {
 	SIZE_T i = 0;
-	while (i < n && src[i]) {
+	while (i < n && src[i] != L'\0') {
 		dest[i] = src[i];
 		i++;
 	}
@@ -568,6 +327,12 @@ NOREGPARM int WRAP_EXPORT(_win_atoi)
 	return i;
 }
 
+STDCALL int WRAP_EXPORT(_win_isprint)
+	(int c)
+{
+	return isprint(c);
+}
+
 STDCALL s64 WRAP_EXPORT(_alldiv)
 	(s64 a, s64 b)
 {
@@ -634,7 +399,7 @@ STDCALL SIZE_T WRAP_EXPORT(RtlCompareMemory)
 	size_t i;
 	char *x, *y;
 
-	TRACEENTER1("%s", "");
+	TRACEENTER1("");
 
 	x = (char *)a;
 	y = (char *)b;
@@ -679,11 +444,10 @@ STDCALL LONG WRAP_EXPORT(RtlCompareString)
 	LONG ret = 0;
 	const char *p1, *p2;
 
-	TRACEENTER1("%s", "");
-	len = min(s1->len, s2->len);
+	TRACEENTER2("");
+	len = min(s1->length, s2->length);
 	p1 = s1->buf;
 	p2 = s2->buf;
-
 	if (case_insensitive)
 		while (!ret && len--)
 			ret = toupper(*p1++) - toupper(*p2++);
@@ -691,8 +455,8 @@ STDCALL LONG WRAP_EXPORT(RtlCompareString)
 		while (!ret && len--)
 			ret = *p1++ - *p2++;
 	if (!ret)
-		ret = s1->len - s2->len;
-	return ret;
+		ret = s1->length - s2->length;
+	TRACEEXIT2(return ret);
 }
 
 STDCALL LONG WRAP_EXPORT(RtlCompareUnicodeString)
@@ -703,29 +467,30 @@ STDCALL LONG WRAP_EXPORT(RtlCompareUnicodeString)
 	LONG ret = 0;
 	const wchar_t *p1, *p2;
 
-	TRACEENTER1("%s", "");
-	len = min(s1->len, s2->len);
+	TRACEENTER2("");
+
+	len = min(s1->length, s2->length) / sizeof(wchar_t);
 	p1 = s1->buf;
 	p2 = s2->buf;
-
 	if (case_insensitive)
 		while (!ret && len--)
 			ret = toupper((u8)*p1++) - toupper((u8)*p2++);
 	else
 		while (!ret && len--)
-			ret = *p1++ - *p2++;
+			ret = (u8)*p1++ - (u8)*p2++;
 	if (!ret)
-		ret = s1->len - s2->len;
-	return ret;
+		ret = s1->length - s2->length;
+	DBGTRACE2("len: %d, ret: %d", len, ret);
+	TRACEEXIT2(return ret);
 }
 
 STDCALL BOOLEAN WRAP_EXPORT(RtlEqualString)
 	(const struct ansi_string *s1, const struct ansi_string *s2,
 	 BOOLEAN case_insensitive)
 {
-	TRACEENTER1("%s", "");
-	if (s1->len != s2->len)
-		return 0;
+	TRACEENTER1("");
+	if (s1->length != s2->length)
+		return FALSE;
 	return !RtlCompareString(s1, s2, case_insensitive);
 }
 
@@ -733,153 +498,170 @@ STDCALL BOOLEAN WRAP_EXPORT(RtlEqualUnicodeString)
 	(const struct unicode_string *s1, const struct unicode_string *s2,
 	 BOOLEAN case_insensitive)
 {
-	if (s1->len != s2->len)
-		return 0;
+	if (s1->length != s2->length)
+		return FALSE;
 	return !RtlCompareUnicodeString(s1, s2, case_insensitive);
 }
 
 STDCALL void WRAP_EXPORT(RtlCopyUnicodeString)
 	(struct unicode_string *dst, struct unicode_string *src)
 {
-	TRACEENTER1("%s", "");
+	TRACEENTER1("%p, %p", dst, src);
 	if (src) {
-		unsigned int len = min(src->len, dst->buflen);
-		memcpy(dst->buf, src->buf, len);
-		dst->len = len;
-		/* append terminating '\0' if enough space */
-		if (len < dst->buflen)
-			dst->buf[len] = 0;
+		dst->length = min(src->length, dst->max_length);
+		memcpy(dst->buf, src->buf, dst->length);
+		if (dst->length < dst->max_length)
+			dst->buf[dst->length / sizeof(wchar_t)] = 0;
 	} else
-		dst->len = 0;
+		dst->max_length = 0;
 	TRACEEXIT1(return);
 }
 
 STDCALL NTSTATUS WRAP_EXPORT(RtlAppendUnicodeToString)
 	(struct unicode_string *dst, wchar_t *src)
 {
-	int i;
-
-	for (i = 0; src[i] != 0 && dst->len + i < dst->buflen; i++)
-		dst->buf[dst->len + i] = src[i];
-	if (src[i] != 0)
-		return STATUS_BUFFER_TOO_SMALL;
-	dst->len += i;
+	if (src) {
+		int len;
+		for (len = 0; src[len]; len++)
+			;
+		if (dst->length + (len * sizeof(wchar_t)) > dst->max_length)
+			return STATUS_BUFFER_TOO_SMALL;
+		memcpy(&dst->buf[dst->length], src, len * sizeof(wchar_t));
+		dst->length += len * sizeof(wchar_t);
+		if (dst->max_length > dst->length)
+			dst->buf[dst->length / sizeof(wchar_t)] = 0;
+	}
 	return STATUS_SUCCESS;
 }
 
 STDCALL NTSTATUS WRAP_EXPORT(RtlAppendUnicodeStringToString)
 	(struct unicode_string *dst, struct unicode_string *src)
 {
-	int i;
-
-	if (dst->buflen < dst->len + src->len)
+	if (dst->max_length < src->length + dst->length)
 		return STATUS_BUFFER_TOO_SMALL;
-
-	for (i = 0; i < src->len; i++)
-		dst->buf[dst->len + i] = src->buf[i];
-	dst->len += i;
-	return STATUS_SUCCESS;
+	if (src->length) {
+		memcpy(&dst->buf[dst->length], src->buf, src->length);
+		dst->length += src->length;
+		if (dst->max_length > dst->length)
+			dst->buf[dst->length / sizeof(wchar_t)] = 0;
+	}
+	TRACEEXIT2(return STATUS_SUCCESS);
 }
 
 STDCALL NTSTATUS WRAP_EXPORT(RtlAnsiStringToUnicodeString)
-	(struct unicode_string *dst, struct ansi_string *src, BOOLEAN dup)
+	(struct unicode_string *dst, const struct ansi_string *src,
+	 BOOLEAN alloc)
 {
 	int i;
-	wchar_t *d;
-	char *s;
 
-	TRACEENTER2("dup: %d src: %s", dup, src->buf);
-	if (dup) {
-		wchar_t *buf = kmalloc((src->buflen+1) * sizeof(wchar_t),
-				       GFP_KERNEL);
-		if (!buf)
-			TRACEEXIT1(return STATUS_FAILURE);
-		dst->buf = buf;
-		dst->buflen = (src->buflen+1) * sizeof(wchar_t);
+	TRACEENTER2("src: length: %d max: %d, buf: %p, dst: %p, alloc: %d",
+		    src->length, src->max_length, src->buf, dst, alloc);
+
+	if (src->buf == NULL || src->length == 0 || src->max_length == 0) {
+		dst->length = 0;
+		if (alloc) {
+			dst->buf = NULL;
+			dst->max_length = 0;
+		} else if (dst->max_length >= sizeof(dst->buf[0]))
+			dst->buf[0] = 0;
+		TRACEEXIT2(return STATUS_SUCCESS);
 	}
-	else if (dst->buflen < (src->len+1) * sizeof(wchar_t))
-		TRACEEXIT1(return STATUS_FAILURE);
+	if (alloc == TRUE) {
+		int len = (src->length + 1) * sizeof(wchar_t);
+		dst->buf = ExAllocatePoolWithTag(NonPagedPool, len, 0);
+		if (!dst->buf) {
+			dst->max_length = dst->length = 0;
+			TRACEEXIT2(return STATUS_NO_MEMORY);
+		}
+		dst->max_length = (src->length + 1) * sizeof(wchar_t);
+	} else if (dst->max_length < (src->length * sizeof(wchar_t)))
+		TRACEEXIT2(return STATUS_BUFFER_TOO_SMALL);
 
-	dst->len = src->len * sizeof(wchar_t);
-	d = dst->buf;
-	s = src->buf;
-	for(i = 0; i < src->len; i++)
-		d[i] = s[i];
-
-	d[i] = 0;
-
-	DBGTRACE2("len = %d", dst->len);
+	dst->length = src->length * sizeof(wchar_t);
+	for(i = 0; i < src->length; i++)
+		dst->buf[i] = src->buf[i];
+	if (dst->max_length > dst->length)
+		dst->buf[dst->length / sizeof(wchar_t)] = 0;
+	DBGTRACE2("dst: length: %d, max_length: %d, string: %s (%p)",
+		  dst->length, dst->max_length, src->buf, src->buf);
 	TRACEEXIT2(return STATUS_SUCCESS);
 }
 
 STDCALL NTSTATUS WRAP_EXPORT(RtlUnicodeStringToAnsiString)
-	(struct ansi_string *dst, struct unicode_string *src, BOOLEAN dup)
+	(struct ansi_string *dst, const struct unicode_string *src,
+	 BOOLEAN alloc)
 {
 	int i;
-	wchar_t *s;
-	char *d;
 
-	TRACEENTER2("dup: %d src->len: %d src->buflen: %d, src->buf: %p,"
-		    "dst: %p", dup, src->len, src->buflen, src->buf, dst);
+	TRACEENTER2("src: length: %d max: %d, buf: %p, dst: %p, alloc: %d",
+		    src->length, src->max_length, src->buf, dst, alloc);
 
-	if (dup) {
-		char *buf = kmalloc((src->buflen+1) / sizeof(wchar_t),
-				    GFP_KERNEL);
-		if (!buf)
-			return STATUS_FAILURE;
-		dst->buf = buf;
-		dst->buflen = (src->buflen+1) / sizeof(wchar_t);
-	} else if (dst->buflen < (src->len+1) / sizeof(wchar_t))
-		return STATUS_FAILURE;
+	if (src->buf == NULL || src->length == 0 || src->max_length == 0) {
+		dst->length = 0;
+		if (alloc) {
+			dst->buf = NULL;
+			dst->max_length = 0;
+		} else if (dst->max_length >= sizeof(dst->buf[0]))
+			dst->buf[0] = 0;
+		TRACEEXIT2(return STATUS_SUCCESS);
+	}
+	if (alloc == TRUE) {
+		int len = (src->length / sizeof(wchar_t)) + 1;
+		dst->buf = ExAllocatePoolWithTag(NonPagedPool, len, 0);
+		if (!dst->buf) {
+			dst->max_length = dst->length = 0;
+			TRACEEXIT2(return STATUS_NO_MEMORY);
+		}
+		dst->max_length = src->length / sizeof(wchar_t) + 1;
+	} else if (dst->max_length < src->length / sizeof(wchar_t))
+		TRACEEXIT2(return STATUS_BUFFER_TOO_SMALL);
 
-	dst->len = src->len / sizeof(wchar_t);
-	s = src->buf;
-	d = dst->buf;
-	for(i = 0; i < dst->len; i++)
-		d[i] = s[i];
-	d[i] = 0;
-
-	DBGTRACE2("len = %d", dst->len);
-	DBGTRACE2("string: %s", dst->buf);
+	dst->length = src->length / sizeof(wchar_t);
+	for(i = 0; i < dst->length; i++)
+		dst->buf[i] = src->buf[i];
+	DBGTRACE2("i: %d", i);
+	if (dst->max_length > dst->length)
+		dst->buf[dst->length] = 0;
+	DBGTRACE2("dst: length: %d, max_length: %d, string: %s",
+		  dst->length, dst->max_length, dst->buf);
 	TRACEEXIT2(return STATUS_SUCCESS);
 }
 
 STDCALL NTSTATUS WRAP_EXPORT(RtlUnicodeStringToInteger)
 	(struct unicode_string *ustring, ULONG base, ULONG *value)
 {
-	int negsign;
+	int i, negsign;
 	wchar_t *str;
 
 	*value = 0;
-	if (ustring->buflen <= 0)
-		return STATUS_INVALID_PARAMETER;
+	if (ustring->length == 0)
+		return STATUS_SUCCESS;
 
 	str = ustring->buf;
-
+	i = 0;
 	negsign = 0;
-	switch ((char)*str) {
+	switch ((char)str[i]) {
 	case '-':
 		negsign = 1;
 		/* fall through */
 	case '+':
-		str++;
+		i++;
 		break;
 	}
 		       
-	if (base == 0 &&
-	    (void *)str < (void *)&ustring->buf[ustring->buflen]) {
-		switch(tolower((char)*str)) {
+	if (base == 0 && i < ustring->length && str[i]) {
+		switch(tolower((char)str[i])) {
 		case 'x':
 			base = 16;
-			str++;
+			i++;
 			break;
 		case 'o':
 			base = 8;
-			str++;
+			i++;
 			break;
 		case 'b':
 			base = 2;
-			str++;
+			i++;
 			break;
 		default:
 			base = 10;
@@ -889,9 +671,9 @@ STDCALL NTSTATUS WRAP_EXPORT(RtlUnicodeStringToInteger)
 	if (!(base == 2 || base == 8 || base == 10 || base == 16))
 		return STATUS_INVALID_PARAMETER;
 
-	for (; (void *)str < (void *)&ustring->buf[ustring->buflen]; str++) {
+	for (; i < ustring->length && str[i]; i++) {
 		int r;
-		char c = tolower((char)*str);
+		char c = tolower((char)str[i]);
 
 		if (c >= '0' && c <= '9')
 			r = c - '0';
@@ -912,11 +694,11 @@ STDCALL NTSTATUS WRAP_EXPORT(RtlUnicodeStringToInteger)
 STDCALL NTSTATUS WRAP_EXPORT(RtlIntegerToUnicodeString)
 	(ULONG value, ULONG base, struct unicode_string *ustring)
 {
-	char string[sizeof(wchar_t) * 8 + 1];
+	char string[sizeof(wchar_t) * (sizeof(ULONG) * 2 + 1)];
 	struct ansi_string ansi;
 	int i;
 
-	TRACEENTER1("%s", "");
+	TRACEENTER2("%u, %u, %p", value, base, ustring);
 	if (base == 0)
 		base = 10;
 	if (!(base == 2 || base == 8 || base == 10 || base == 16))
@@ -937,50 +719,56 @@ STDCALL NTSTATUS WRAP_EXPORT(RtlIntegerToUnicodeString)
 		return STATUS_BUFFER_TOO_SMALL;
 
 	ansi.buf = string;
-	ansi.len = strlen(string);
-	ansi.buflen = sizeof(string);
-	return RtlAnsiStringToUnicodeString(ustring, &ansi, 0);
+	ansi.length = strlen(string);
+	ansi.max_length = sizeof(string);
+	return RtlAnsiStringToUnicodeString(ustring, &ansi, FALSE);
 }
 
 STDCALL void WRAP_EXPORT(RtlInitUnicodeString)
-	(struct unicode_string *dest, const wchar_t *src)
+	(struct unicode_string *dst, const wchar_t *src)
 {
-	TRACEENTER1("%s", "");
-	if (dest == NULL)
+	TRACEENTER2("%p", dst);
+	if (dst == NULL)
 		TRACEEXIT1(return);
 	if (src == NULL) {
-		dest->len = dest->buflen = 0;
-		dest->buf = NULL;
+		dst->max_length = dst->length = 0;
+		dst->buf = NULL;
 	} else {
 		int i = 0;
 		while (src[i])
 			i++;
-		dest->buf = (wchar_t *)src;
-		dest->len = dest->buflen = i * 2;
+		dst->buf = (wchar_t *)src;
+		dst->length = i * sizeof(wchar_t);
+		dst->max_length = (i + 1) * sizeof(wchar_t);
 	}
 	TRACEEXIT1(return);
 }
 
 STDCALL void WRAP_EXPORT(RtlInitAnsiString)
-	(struct ansi_string *dst, CHAR *src)
+	(struct ansi_string *dst, const char *src)
 {
-	TRACEENTER2("%s", "");
+	TRACEENTER2("%p", dst);
 	if (dst == NULL)
 		TRACEEXIT2(return);
 	if (src == NULL) {
-		dst->len = dst->buflen = 0;
+		dst->max_length = dst->length = 0;
 		dst->buf = NULL;
-		TRACEEXIT2(return);
+	} else {
+		int i = 0;
+		while (src[i])
+			i++;
+		dst->buf = (char *)src;
+		dst->length = i;
+		dst->max_length = i + 1;
 	}
-	dst->len = dst->buflen = strlen(src);
-	dst->buf = (char *)src;
+	DBGTRACE2("%p", dst->buf);
 	TRACEEXIT2(return);
 }
 
 STDCALL void WRAP_EXPORT(RtlInitString)
-	(struct ansi_string *dst, CHAR *src)
+	(struct ansi_string *dst, const char *src)
 {
-	TRACEENTER2("%s", "");
+	TRACEENTER2("%p", dst);
 	RtlInitAnsiString(dst, src);
 	TRACEEXIT2(return);
 }
@@ -988,11 +776,13 @@ STDCALL void WRAP_EXPORT(RtlInitString)
 STDCALL void WRAP_EXPORT(RtlFreeUnicodeString)
 	(struct unicode_string *string)
 {
-	if (string == NULL || string->buf == NULL)
+	TRACEENTER2("%p", string);
+	if (string == NULL)
 		return;
 
-	kfree(string->buf);
-	string->buflen = string->len = 0;
+	if (string->buf)
+		ExFreePool(string->buf);
+	string->length = string->max_length = 0;
 	string->buf = NULL;
 	return;
 }
@@ -1000,28 +790,119 @@ STDCALL void WRAP_EXPORT(RtlFreeUnicodeString)
 STDCALL void WRAP_EXPORT(RtlFreeAnsiString)
 	(struct ansi_string *string)
 {
-	if (string == NULL || string->buf == NULL)
+	TRACEENTER2("%p", string);
+	if (string == NULL)
 		return;
-
-	kfree(string->buf);
-	string->buflen = string->len = 0;
+	if (string->buf)
+		ExFreePool(string->buf);
+	string->length = string->max_length = 0;
 	string->buf = NULL;
 	return;
 }
 
 STDCALL NTSTATUS WRAP_EXPORT(RtlQueryRegistryValues)
-	(ULONG relative, const wchar_t *path, void *tbl, void *context,
-	 void *env)
+	(ULONG relative, wchar_t *path, struct rtl_query_registry_table *tbl,
+	 void *context, void *env)
 {
-	TRACEENTER5("%s", "");
+	struct ansi_string ansi;
+	struct unicode_string unicode;
+	NTSTATUS status, ret;
+
+	TRACEENTER3("%x, %p", relative, tbl);
 	UNIMPL();
-	TRACEEXIT5(return STATUS_SUCCESS);
+
+	unicode.buf = path;
+	unicode.length = _win_wcslen(path) * sizeof(wchar_t);
+	unicode.max_length = unicode.length + sizeof(wchar_t);
+	if (RtlUnicodeStringToAnsiString(&ansi, &unicode, TRUE) ==
+	    STATUS_SUCCESS) {
+		DBGTRACE2("%s", ansi.buf);
+		RtlFreeAnsiString(&ansi);
+	}
+	ret = STATUS_SUCCESS;
+	for (; tbl->name; tbl++) {
+		unicode.buf = tbl->name;
+		unicode.length = _win_wcslen(tbl->name) * sizeof(wchar_t);
+		unicode.max_length = unicode.length + sizeof(wchar_t);
+		if (RtlUnicodeStringToAnsiString(&ansi, &unicode, TRUE) ==
+		    STATUS_SUCCESS) {
+			DBGTRACE2("name: %s", ansi.buf);
+			RtlFreeAnsiString(&ansi);
+		}
+		DBGTRACE2("flags: %08X", tbl->flags);
+		if (tbl->flags == RTL_QUERY_REGISTRY_DIRECT) {
+			DBGTRACE2("type: %08X", tbl->def_type);
+			if (tbl->def_type == REG_DWORD) {
+				/* Atheros USB driver needs this, but
+				 * don't know where and how to get its
+				 * value */
+				if (tbl->def_data) {
+					DBGTRACE2("def_data: %x",
+						  *(int *)tbl->def_data);
+					*(DWORD *)tbl->context =
+						*(DWORD *)tbl->def_data;
+				} else
+					*(DWORD *)tbl->context = 0x2345dbe;
+			}
+		} else {
+			void *data;
+			ULONG type, length;
+
+			if (!tbl->query_func) {
+				ERROR("oops: no query_func");
+				ret = STATUS_INVALID_PARAMETER;
+				break;
+			}
+			if (tbl->flags & RTL_QUERY_REGISTRY_NOVALUE) {
+				data = NULL;
+				type = REG_NONE;
+				length = 0;
+			} else {
+				data = tbl->def_data;
+				type = tbl->def_type;
+				length = tbl->def_length;;
+			}
+			DBGTRACE2("calling query_func: %p", tbl->query_func);
+			status = LIN2WIN6(tbl->query_func, tbl->name, type,
+					  data, length, context, env);
+			DBGTRACE2("status: %08X", status);
+			if (status) {
+				if (status == STATUS_BUFFER_TOO_SMALL)
+					ret = STATUS_BUFFER_TOO_SMALL;
+				else
+					TRACEEXIT2(return STATUS_INVALID_PARAMETER);
+			}
+		}
+	}
+	TRACEEXIT3(return ret);
 }
 
 STDCALL NTSTATUS WRAP_EXPORT(RtlWriteRegistryValue)
-	(ULONG relative, const wchar_t *path, const wchar_t *name, ULONG type,
+	(ULONG relative, wchar_t *path, wchar_t *name, ULONG type,
 	 void *data, ULONG length)
 {
+	struct ansi_string ansi;
+	struct unicode_string unicode;
+
+	TRACEENTER3("%d", relative);
+	UNIMPL();
+
+	unicode.buf = path;
+	unicode.length = _win_wcslen(path) * sizeof(wchar_t);
+	unicode.max_length = unicode.length + sizeof(wchar_t);
+	if (RtlUnicodeStringToAnsiString(&ansi, &unicode, TRUE) ==
+	    STATUS_SUCCESS) {
+		DBGTRACE2("%s", ansi.buf);
+		RtlFreeAnsiString(&ansi);
+	}
+	unicode.buf = name;
+	unicode.length = _win_wcslen(name) * sizeof(wchar_t);
+	unicode.max_length = unicode.length + sizeof(wchar_t);
+	if (RtlUnicodeStringToAnsiString(&ansi, &unicode, TRUE) ==
+	    STATUS_SUCCESS) {
+		DBGTRACE2("%s", ansi.buf);
+		RtlFreeAnsiString(&ansi);
+	}
 	TRACEEXIT5(return STATUS_SUCCESS);
 }
 
@@ -1052,17 +933,6 @@ STDCALL int WRAP_EXPORT(rand)
 	return r;
 }
 
-ULONGLONG ticks_1601(void)
-{
-	struct timeval now;
-	ULONGLONG ticks;
-
-	do_gettimeofday(&now);
-	ticks = (ULONGLONG)now.tv_sec * TICKSPERSEC;
-	ticks += now.tv_usec * 10 + TICKS_1601_TO_1970;
-	return ticks;
-}
-
 void WRAP_EXPORT(RtlUnwind)(void){UNIMPL();}
 
 int stricmp(const char *s1, const char *s2)
@@ -1094,21 +964,22 @@ void dump_stack(void)
 		printk(KERN_DEBUG "sp[%d] = %p\n", i, (void *)sp[i]);
 }
 
-void dump_bytes(const char *where, const u8 *ip)
+void dump_bytes(const char *name, const u8 *from, int len)
 {
 	int i, j;
-	u8 code[50];
+	u8 code[100];
 
 	memset(code, 0, sizeof(code));
-	for (i = j = 0; i < 16; i++, j += 3) {
-		if (j+3 > sizeof(code))
+	for (i = j = 0; i < len; i++, j += 3) {
+		if (j+3 > sizeof(code)) {
 			ERROR("not enough space: %u > %u", j+3,
 			      (unsigned int)sizeof(code));
-		else
-			sprintf(&code[j], "%02x ", ip[i]);
+			break;
+		} else
+			sprintf(&code[j], "%02x ", from[i]);
 	}
 	code[sizeof(code)-1] = 0;
-	printk(KERN_DEBUG "%s: %p: %s\n", where, ip, code);
+	printk(KERN_DEBUG "%s: %p: %s\n", name, from, code);
 }
 
 #include "misc_funcs_exports.h"
