@@ -71,6 +71,8 @@ struct ndis_buffer_pool {
 #define fPACKET_CONTAINS_MEDIA_SPECIFIC_INFO	0x40
 #define fPACKET_ALLOCATED_BY_NDIS		0x80
 
+#define PROTOCOL_RESERVED_SIZE_IN_PACKET (4 * sizeof(void *))
+
 enum ndis_per_packet_info {
 	TcpIpChecksumPacketInfo, IpSecPacketInfo, TcpLargeSendPacketInfo,
 	ClassificationHandlePacketInfo, NdisReserved,
@@ -108,10 +110,30 @@ struct ndis_packet_oob_data {
 	NDIS_STATUS status;
 };
 
+/* ndiswrapper specific info */
+struct wrap_ndis_packet {
+	struct ndis_packet_oob_data oob_data;
+	struct ndis_packet_extension extension;
+
+	struct ndis_packet *next;
+	struct scatterlist *sg_list;
+	unsigned int sg_ents;
+	struct ndis_sg_list ndis_sg_list;
+	struct ndis_sg_element *ndis_sg_elements;
+	/* RTL8180L overshoots past ndis_eg_elements (during
+	 * MiniportSendPackets) and overwrites what is below, if SG
+	 * DMA is used, so don't use ndis_sg_element in that
+	 * case. This structure is used only when SG is disabled */
+	struct ndis_sg_element ndis_sg_element;
+
+	unsigned char header[ETH_HLEN];
+	unsigned char *look_ahead;
+	UINT look_ahead_size;
+	struct sk_buff *skb;
+};
+
 struct ndis_packet {
-
 	struct ndis_packet_private private;
-
 	/* for use by miniport */
 	union {
 		/* for connectionless mininports */
@@ -129,39 +151,19 @@ struct ndis_packet {
 		} mac_reserved;
 	} u;
 	ULONG_PTR reserved[2];
+//	UCHAR protocol_reserved[PROTOCOL_RESERVED_SIZE_IN_PACKET];
 	UCHAR protocol_reserved[1];
-
-	struct ndis_packet_oob_data oob_data;
-	struct ndis_packet_extension extension;
-
-	/* ndiswrapper specific info */
-	struct ndis_packet *next;
-	struct scatterlist *sg_list;
-	unsigned int sg_ents;
-	/* ndiswrapper-specific info */
-	struct ndis_sg_list ndis_sg_list;
-	struct ndis_sg_element *ndis_sg_elements;
-	/* RTL8180L overshoots past ndis_eg_elements (during
-	 * MiniportSendPackets) and overwrites what is below, if SG
-	 * DMA is used, so don't use ndis_sg_element in that
-	 * case. This structure is used only when SG is disabled */
-	struct ndis_sg_element ndis_sg_element;
-
-	unsigned char header[ETH_HLEN];
-	unsigned char *look_ahead;
-	UINT look_ahead_size;
-};
-
-struct wrap_ndis_packet {
-	struct nt_list list;
-	struct ndis_packet ndis_packet;
+	struct wrap_ndis_packet *wrap_ndis_packet;
 };
 
 struct ndis_packet_pool {
-	int max_descr;
-	int num_allocated_descr;
+	UINT max_descr;
+	UINT num_allocated_descr;
+	UINT num_used_descr;
 	struct ndis_packet *free_descr;
 	KSPIN_LOCK lock;
+	UINT proto_rsvd_length;
+	struct nt_list list;
 };
 
 enum ndis_device_pnp_event {
@@ -222,14 +224,18 @@ enum ndis_pm_state {
 	NdisDeviceStateD3, NdisDeviceStateMaximum
 };
 
+enum ndis_power_profile {
+	NdisPowerProfileBattery, NdisPowerProfileAcOnLine
+};
+
 typedef void (*ndis_isr_handler)(unsigned int *taken, unsigned int *callme,
 				 void *ctx) STDCALL;
 typedef void (*ndis_interrupt_handler)(void *ctx) STDCALL;
 
 struct miniport_char {
 	/* NDIS 3.0 */
-	UCHAR majorVersion;
-	UCHAR minorVersion;
+	UCHAR major_version;
+	UCHAR minor_version;
 	USHORT filler;
 	UINT reserved;
 
@@ -238,45 +244,48 @@ struct miniport_char {
 	void (*enable_interrupts)(void *ctx) STDCALL;
 
 	/* Stop miniport */
-	void (*halt)(void *ctx) STDCALL;
+	void (*miniport_halt)(void *ctx) STDCALL;
 
 	/* Interrupt BH */
 	ndis_interrupt_handler handle_interrupt;
 
 	/* Start miniport driver */
-	UINT (*init)(UINT *status, UINT *medium_index,
-		     enum ndis_medium medium[], UINT medium_array_size,
-		     void *handle, void *conf_handle) STDCALL;
+	NDIS_STATUS (*init)(NDIS_STATUS *error_status, UINT *medium_index,
+			    enum ndis_medium medium[], UINT medium_array_size,
+			    void *handle, void *conf_handle) STDCALL;
 
 	/* Interrupt TH */
 	ndis_isr_handler isr;
 
 	/* Query parameters */
-	UINT (*query)(void *ctx, UINT oid, char *buffer,
-		      UINT buflen, UINT *written, UINT *needed) STDCALL;
+	NDIS_STATUS (*query)(void *ctx, ndis_oid oid, void *buffer,
+			     ULONG buflen, ULONG *written,
+			     ULONG *needed) STDCALL;
 
 	void *reconfig;
-	INT (*reset)(INT *needs_set, void *ctx) STDCALL;
+	NDIS_STATUS (*reset)(BOOLEAN *reset_address, void *ctx) STDCALL;
 
 	/* Send one packet */
-	UINT (*send)(void *ctx, struct ndis_packet *packet,
-		     UINT flags) STDCALL;
+	NDIS_STATUS (*send)(void *ctx, struct ndis_packet *packet,
+			    UINT flags) STDCALL;
 
 	/* Set parameters */
-	UINT (*setinfo)(void *ctx, UINT oid, char *buffer, UINT buflen,
-			UINT *written, UINT *needed) STDCALL;
+	NDIS_STATUS (*setinfo)(void *ctx, ndis_oid oid, void *buffer,
+			       ULONG buflen, ULONG *written,
+			       ULONG *needed) STDCALL;
 
 	/* transfer data from received packet */
-	UINT (*tx_data)(struct ndis_packet *ndis_packet, UINT *bytes_txed,
-				void *adapter_ctx, void *rx_ctx,
-				UINT offset, UINT bytes_to_tx) STDCALL;
+	NDIS_STATUS (*tx_data)(struct ndis_packet *ndis_packet,
+			       UINT *bytes_txed, void *adapter_ctx,
+			       void *rx_ctx, UINT offset,
+			       UINT bytes_to_tx) STDCALL;
 
 	/* NDIS 4.0 extensions */
 	/* upper layer is done with RX packet */
 	void (*return_packet)(void *ctx, void *packet) STDCALL;
 
 	/* Send packets */
-	UINT (*send_packets)(void *ctx, struct ndis_packet **packets,
+	void (*send_packets)(void *ctx, struct ndis_packet **packets,
 			     INT nr_of_packets) STDCALL;
 
 	void (*alloc_complete)(void *handle, void *virt,
@@ -284,17 +293,18 @@ struct miniport_char {
 			       ULONG size, void *ctx) STDCALL;
 
 	/* NDIS 5.0 extensions */
-	UINT (*co_create_vc)(void *ctx, void *vc_handle, void *vc_ctx) STDCALL;
-	UINT (*co_delete_vc)(void *vc_ctx) STDCALL;
-	UINT (*co_activate_vc)(void *vc_ctx, void *call_params) STDCALL;
-	UINT (*co_deactivate_vc)(void *vc_ctx) STDCALL;
-	UINT (*co_send_packets)(void *vc_ctx, void **packets,
-				UINT nr_of_packets) STDCALL;
-	UINT (*co_request)(void *ctx, void *vc_ctx, UINT *req) STDCALL;
+	NDIS_STATUS (*co_create_vc)(void *ctx, void *vc_handle,
+				    void *vc_ctx) STDCALL;
+	NDIS_STATUS (*co_delete_vc)(void *vc_ctx) STDCALL;
+	NDIS_STATUS (*co_activate_vc)(void *vc_ctx, void *call_params) STDCALL;
+	NDIS_STATUS (*co_deactivate_vc)(void *vc_ctx) STDCALL;
+	NDIS_STATUS (*co_send_packets)(void *vc_ctx, void **packets,
+				       UINT nr_of_packets) STDCALL;
+	NDIS_STATUS (*co_request)(void *ctx, void *vc_ctx, UINT *req) STDCALL;
 
 	/* NDIS 5.1 extensions */
-	void *cancel_send_packets;
-	void (*pnp_event_notify)(void *ctx, enum ndis_device_pnp_event,
+	void (*cancel_send_packets)(void *ctx, void *id) STDCALL;
+	void (*pnp_event_notify)(void *ctx, enum ndis_device_pnp_event event,
 				 void *inf_buf, ULONG inf_buf_len) STDCALL;
 	void (*adapter_shutdown)(void *ctx) STDCALL;
 	void *reserved1;
@@ -322,13 +332,7 @@ struct ndis_rw_lock {
 		} s;
 		UCHAR reserved[16];
 	} u;
-    union ndis_rw_lock_refcount ref_count[MAXIMUM_PROCESSORS];
-};
-
-struct wd_ctx_entry {
-	struct list_head list;
-	void *wd;
-	void *ctx;
+	union ndis_rw_lock_refcount ref_count[MAXIMUM_PROCESSORS];
 };
 
 struct ndis_sched_work_item {
@@ -337,12 +341,6 @@ struct ndis_sched_work_item {
 	UCHAR reserved[8 * sizeof(void *)];
 };
 
-struct ndis_io_work_item {
-	void *ctx;
-	void *device_object;
-	void (*func)(void *device_object, void *ctx) STDCALL;
-};
-
 struct ndis_alloc_mem_work_item {
 	unsigned long size;
 	char cached;
@@ -351,27 +349,18 @@ struct ndis_alloc_mem_work_item {
 
 struct ndis_free_mem_work_item {
 	void *addr;
-	unsigned int length;
-	unsigned int flags;
 };
 
-enum ndis_work_entry_type {
-	NDIS_SCHED_WORK_ITEM, NDIS_ALLOC_MEM_WORK_ITEM,
-	NDIS_FREE_MEM_WORK_ITEM, NDIS_IO_WORK_ITEM,
-	NDIS_RETURN_PACKET_WORK_ITEM,
+struct ndis_work_item {
+	void *context;
+	void *routine;
+	UCHAR reserved[8 * sizeof(void *)];
 };
 
-struct ndis_work_entry {
-	struct list_head list;
-	enum ndis_work_entry_type type;
-	struct wrapper_dev *wd;
-	union {
-		struct ndis_sched_work_item *sched_work_item;
-		struct ndis_alloc_mem_work_item alloc_mem_work_item;
-		struct ndis_free_mem_work_item free_mem_work_item;
-		struct ndis_io_work_item *io_work_item;
-		struct ndis_packet *return_packet;
-	} entry;
+struct alloc_shared_mem {
+	void *ctx;
+	ULONG size;
+	BOOLEAN cached;
 };
 
 struct ndis_irq {
@@ -385,13 +374,12 @@ struct ndis_irq {
 	void *id;
 	ndis_isr_handler isr;
 	void *dpc;
-
 	struct kdpc intr_dpc;
 	struct wrapper_dev *wd;
 	UCHAR dpc_count;
-	/* unsigned char filler1 is used for enabled */
-	UCHAR enabled;
-	struct kevent completed_event;
+	/* unsigned char filler1 is used for pending */
+	UCHAR pending;
+	struct nt_event completed_event;
 	UCHAR shared;
 	UCHAR req_isr;
 };
@@ -417,7 +405,7 @@ struct ndis_config_param {
 };
 
 struct device_setting {
-	struct list_head list;
+	struct nt_list list;
 	char name[MAX_NDIS_SETTING_NAME_LEN];
 	char value[MAX_NDIS_SETTING_VALUE_LEN];
 	struct ndis_config_param config_param;
@@ -434,7 +422,7 @@ struct ndis_bin_file {
 #define CE_MINIPORT_CLIENT_ID 2
 
 struct ndis_driver {
-	struct list_head list;
+	struct nt_list list;
 	struct driver_object *drv_obj;
 	char name[MAX_NDIS_SETTING_NAME_LEN];
 	char version[MAX_NDIS_SETTING_VALUE_LEN];
@@ -450,7 +438,7 @@ struct ndis_driver {
 struct wrapper_dev;
 
 struct ndis_device {
-	struct list_head settings;
+	struct nt_list settings;
 	int bustype;
 	int vendor;
 	int device;
@@ -464,7 +452,7 @@ struct ndis_device {
 };
 
 struct ndis_wireless_stats {
-	LARGE_INTEGER length;
+	ULONG length;
 	LARGE_INTEGER tx_frag;
 	LARGE_INTEGER tx_multi_frag;
 	LARGE_INTEGER failed;
@@ -477,6 +465,18 @@ struct ndis_wireless_stats {
 	LARGE_INTEGER rx_frag;
 	LARGE_INTEGER rx_multi_frag;
 	LARGE_INTEGER fcs_err;
+	LARGE_INTEGER tkip_local_mic_failures;
+	LARGE_INTEGER tkip_icv_errors;
+	LARGE_INTEGER tkip_counter_measures_invoked;
+	LARGE_INTEGER tkip_replays;
+	LARGE_INTEGER ccmp_format_errors;
+	LARGE_INTEGER ccmp_replays;
+	LARGE_INTEGER ccmp_decrypt_errors;
+	LARGE_INTEGER fourway_handshake_failures;
+	LARGE_INTEGER wep_undecryptable_count;
+	LARGE_INTEGER wep_icv_errorcount;
+	LARGE_INTEGER decrypt_success_count;
+	LARGE_INTEGER decrypt_failure_count;
 };
 
 enum ndis_status_type {
@@ -486,8 +486,7 @@ enum ndis_status_type {
 	Ndis802_11StatusType_RadioState,
 };
 
-struct ndis_status_indication
-{
+struct ndis_status_indication {
 	enum ndis_status_type status_type;
 };
 
@@ -503,18 +502,19 @@ struct ndis_radio_status_indication
 };
 
 enum wrapper_work {
-	LINK_STATUS_CHANGED, SET_INFRA_MODE, SET_ESSID, SET_PACKET_FILTER,
+	LINK_STATUS_CHANGED, SET_INFRA_MODE, SET_ESSID, SET_MULTICAST_LIST,
 	COLLECT_STATS, SUSPEND_RESUME, HANGCHECK,
 	/* do not work when this is set */
 	SHUTDOWN
 };
 
 enum ndis_attributes {
-	ATTR_SERIALIZED, ATTR_SURPRISE_REMOVE, ATTR_HALT_ON_SUSPEND,
+	ATTR_SERIALIZED, ATTR_SURPRISE_REMOVE, ATTR_NO_HALT_ON_SUSPEND,
 };
 
 enum hw_status {
-	HW_NORMAL, HW_SUSPENDED, HW_HALTED, HW_UNAVAILABLE,
+	HW_NORMAL, HW_SUSPENDED, HW_HALTED, HW_RMMOD, HW_AVAILABLE,
+	HW_INITIALIZED,
 };
 
 struct encr_info {
@@ -556,17 +556,42 @@ enum encryption_status {
 	Ndis802_11Encryption3Enabled, Ndis802_11Encryption3KeyAbsent
 };
 
+struct ndis_auth_encr_pair {
+	enum authentication_mode auth_mode;
+	enum encryption_status encr_mode;
+};
+
+struct ndis_capability {
+	ULONG length;
+	ULONG version;
+	ULONG num_PMKIDs;
+	ULONG num_auth_encr_pair;
+	struct ndis_auth_encr_pair auth_encr_pair[1];
+};
+
+struct ndis_guid {
+	struct guid guid;
+	union {
+		ndis_oid oid;
+		NDIS_STATUS status;
+	};
+	ULONG size;
+	ULONG flags;
+};
+
 struct ndis_timer {
-	struct ktimer ktimer;
+	struct nt_timer nt_timer;
 	struct kdpc kdpc;
 };
 
 struct ndis_miniport_timer {
-	struct ktimer ktimer;
+	struct nt_timer nt_timer;
 	struct kdpc kdpc;
-	void *timer_func;
-	void *timer_ctx;
-	struct wrapper_dev *wd;
+	/* since kdpc already can store func, ctx, I don't know what
+	 * the following two fields are for */
+	DPC func;
+	void *ctx;
+	void *nmb;
 	struct ndis_miniport_timer *next;
 };
 
@@ -626,7 +651,7 @@ struct ndis_resource_list {
 #define MAX_NDIS_PCI_RESOURCES 20
 
 struct ndis_event {
-	struct kevent kevent;
+	struct nt_event nt_event;
 };
 
 struct ndis_bind_paths {
@@ -652,7 +677,7 @@ struct ndis_miniport_interrupt {
 	struct ndis_miniport_block *miniport;
 	UCHAR dpc_count;
 	BOOLEAN filler1;
-	struct kevent dpcs_completed_event;
+	struct nt_event dpcs_completed_event;
         BOOLEAN shared_interrupt;
 	BOOLEAN isr_requested;
 };
@@ -661,10 +686,10 @@ struct ndis_filterdbs {
 	union {
 		void *eth_db;
 		void *null_db;
-	} u;
-	void *trdb;
-	void *fddidb;
-	void *arcdb;
+	};
+	void *tr_db;
+	void *fddi_db;
+	void *arc_db;
 };
 
 enum ndis_interface_type {
@@ -673,6 +698,11 @@ enum ndis_interface_type {
 	NdisInterfacePcMcia,
 };
 
+struct auth_encr_capa {
+	unsigned long auth;
+	unsigned long encr;
+};
+
 /*
  * This struct contains function pointers that the drivers references
  * directly via macros, so it's important that they are at the correct
@@ -774,19 +804,17 @@ struct wrapper_dev {
 	struct ndis_miniport_block *nmb;
 	struct ndis_driver *driver;
 	struct phys_dev dev;
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
-	struct usb_interface *intf;
-#endif
 	struct net_device *net_dev;
 	void *shutdown_ctx;
 
-	struct work_struct irq_work;
+	struct tasklet_struct irq_tasklet;
 	struct ndis_irq *ndis_irq;
 	unsigned long mem_start;
 	unsigned long mem_end;
 
 	struct net_device_stats stats;
 	struct iw_statistics wireless_stats;
+	BOOLEAN stats_enabled;
 	struct ndis_wireless_stats ndis_stats;
 	struct ndis_device *ndis_device;
 
@@ -803,20 +831,17 @@ struct wrapper_dev {
 
 	struct semaphore ndis_comm_mutex;
 	wait_queue_head_t ndis_comm_wq;
-	int ndis_comm_res;
 	int ndis_comm_done;
+	int ndis_comm_status;
+	ULONG packet_filter;
 
 	int serialized;
 	int use_sg_dma;
 	int map_count;
-	int multicast_list_size;
-	char *multicast_list;
 	dma_addr_t *map_dma_addr;
 
 	int hangcheck_interval;
-	int hangcheck_active;
 	struct timer_list hangcheck_timer;
-	int reset_status;
 
 	struct timer_list stats_timer;
 
@@ -831,15 +856,16 @@ struct wrapper_dev {
 
 	struct ndis_essid essid;
 
-	unsigned long capa;
+	struct auth_encr_capa capa;
 	enum authentication_mode auth_mode;
 	enum encryption_status encr_mode;
 	enum network_infrastructure infrastructure_mode;
+	int num_pmkids;
 
 	mac_address mac;
 
 	/* list of initialized timers */
-	struct nt_list wrapper_timer_list;
+	struct nt_list wrap_timer_list;
 	KSPIN_LOCK timer_lock;
 
 	struct proc_dir_entry *procfs_iface;
@@ -856,23 +882,37 @@ struct wrapper_dev {
 	int iw_auth_cipher_group;
 	int iw_auth_key_mgmt;
 	int iw_auth_80211_auth_alg;
+	struct ndis_packet_pool *wrapper_packet_pool;
+	struct ndis_buffer_pool *wrapper_buffer_pool;
 };
 
-struct ndis_pmkid_candidate
-{
+struct ndis_pmkid_candidate {
 	mac_address bssid;
 	unsigned long flags;
 };
 
-struct ndis_pmkid_candidate_list
-{
+struct ndis_pmkid_candidate_list {
 	unsigned long version;
 	unsigned long num_candidates;
 	struct ndis_pmkid_candidate candidates[1];
 };
 
-struct ndis_packet *allocate_ndis_packet(void);
-void free_ndis_packet(struct ndis_packet *packet);
+STDCALL void NdisAllocatePacketPoolEx
+	(NDIS_STATUS *status, struct ndis_packet_pool **pool_handle,
+	 UINT num_descr, UINT overflowsize, UINT proto_rsvd_length);
+STDCALL void NdisFreePacketPool(struct ndis_packet_pool *pool);
+STDCALL void NdisAllocatePacket(NDIS_STATUS *status,
+				struct ndis_packet **packet,
+				struct ndis_packet_pool *pool);
+STDCALL void NdisFreePacket(struct ndis_packet *descr);
+STDCALL void NdisAllocateBufferPool
+	(NDIS_STATUS *status, struct ndis_buffer_pool **pool_handle,
+	 UINT num_descr);
+STDCALL void NdisFreeBufferPool(struct ndis_buffer_pool *pool);
+STDCALL void NdisAllocateBuffer
+	(NDIS_STATUS *status, ndis_buffer **buffer,
+	 struct ndis_buffer_pool *pool, void *virt, UINT length);
+STDCALL void NdisFreeBuffer(ndis_buffer *descr);
 STDCALL void NdisMIndicateReceivePacket(struct ndis_miniport_block *nmb,
 					struct ndis_packet **packets,
 					UINT nr_packets);
@@ -892,8 +932,8 @@ STDCALL ULONG NDIS_BUFFER_TO_SPAN_PAGES(ndis_buffer *buffer);
 STDCALL BOOLEAN NdisWaitEvent(struct ndis_event *event, UINT timeout);
 STDCALL void NdisSetEvent(struct ndis_event *event);
 STDCALL void NdisMDeregisterInterrupt(struct ndis_irq *ndis_irq);
-STDCALL void EthRxIndicateHandler(void *adapter_ctx, void *rx_ctx,
-				  char *header1, char *header,
+STDCALL void EthRxIndicateHandler(struct ndis_miniport_block *nmb,
+				  void *rx_ctx, char *header1, char *header,
 				  UINT header_size, void *look_ahead,
 				  UINT look_ahead_size, UINT packet_size);
 STDCALL void EthRxComplete(struct ndis_miniport_block *nmb);
@@ -905,14 +945,24 @@ STDCALL void NdisWriteConfiguration(NDIS_STATUS *status,
 				    struct unicode_string *key,
 				    struct ndis_config_param *val);
 
+STDCALL void NdisReadConfiguration
+	(NDIS_STATUS *status, struct ndis_config_param **dest,
+	 struct ndis_miniport_block *nmb, struct unicode_string *key,
+	 enum ndis_config_param_type type);
+
 void *get_sp(void);
 int ndis_init(void);
 void ndis_exit_device(struct wrapper_dev *wd);
 void ndis_exit(void);
+void insert_ndis_kdpc_work(struct kdpc *kdpc);
+BOOLEAN remove_ndis_kdpc_work(struct kdpc *kdpc);
 
 int usb_init(void);
 void usb_exit(void);
+int usb_init_device(struct wrapper_dev *wd);
+void usb_exit_device(struct wrapper_dev *wd);
 void usb_cleanup(void);
+void usb_cancel_pending_urbs(void);
 
 int load_pe_images(struct pe_image[], int n);
 
@@ -922,12 +972,15 @@ void ndiswrapper_procfs_remove_iface(struct wrapper_dev *wd);
 void ndiswrapper_procfs_remove(void);
 
 int misc_funcs_init(void);
+int misc_funcs_init_device(struct wrapper_dev *wd);
 void misc_funcs_exit_device(struct wrapper_dev *wd);
 void misc_funcs_exit(void);
 
-void packet_recycler(void *param);
 int stricmp(const char *s1, const char *s2);
-void dump_bytes(const char *where, const u8 *ip);
+void dump_bytes(const char *name, const u8 *from, int len);
+
+int pdo_suspend_pci(struct pci_dev *pdev, pm_message_t state);
+int pdo_resume_pci(struct pci_dev *pdev);
 
 /* Required OIDs */
 #define OID_GEN_SUPPORTED_LIST			0x00010101
@@ -1209,4 +1262,7 @@ void dump_bytes(const char *where, const u8 *ip);
 #define NDIS_ATTRIBUTE_NOT_CO_NDIS              0x00000100
 #define NDIS_ATTRIBUTE_USES_SAFE_BUFFER_APIS    0x00000200
 
+#define NDIS_FLAGS_PROTOCOL_ID_MASK		0x0000000F
+#define NDIS_PROTOCOL_ID_TCP_IP			0x02
+
 #endif /* NDIS_H */
