@@ -16,6 +16,8 @@
 #ifndef _NTOSKERNEL_H_
 #define _NTOSKERNEL_H_
 
+#define UTILS_VERSION "1.5"
+
 #include <linux/types.h>
 #include <linux/timer.h>
 #include <linux/time.h>
@@ -29,13 +31,18 @@
 #include <linux/mm.h>
 #include <linux/random.h>
 #include <linux/ctype.h>
-#include <linux/usb.h>
 #include <linux/list.h>
-
+#include <linux/sched.h>
+#include <linux/usb.h>
 #include <linux/spinlock.h>
 #include <asm/mman.h>
-
 #include <linux/version.h>
+#include <linux/etherdevice.h>
+#include <net/iw_handler.h>
+#include <linux/netdevice.h>
+#include <linux/ethtool.h>
+#include <linux/if_arp.h>
+#include <linux/rtnetlink.h>
 
 #include "winnt_types.h"
 #include "ndiswrapper.h"
@@ -45,6 +52,11 @@
 #include <linux/kthread.h>
 #endif
 
+#if defined(DISABLE_USB)
+#undef CONFIG_USB
+#undef CONFIG_USB_MODULE
+#endif
+
 #if !defined(CONFIG_USB) && defined(CONFIG_USB_MODULE)
 #define CONFIG_USB 1
 #endif
@@ -152,6 +164,10 @@ typedef task_queue workqueue;
 #define preempt_disable() do { } while (0)
 #endif
 
+#ifndef preempt_enable_no_resched
+#define preempt_enable_no_resched() preempt_enable()
+#endif
+
 #ifndef container_of
 #define container_of(ptr, type, member)					\
 ({									\
@@ -170,12 +186,43 @@ typedef task_queue workqueue;
 
 #endif // LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 
+#ifndef PMSG_SUSPEND
+#ifdef PM_SUSPEND
+/* this is not correct - the value of PM_SUSPEND is different from
+ * PMSG_SUSPEND, but ndiswrapper doesn't care about the value when
+ * suspending */
+#define PMSG_SUSPEND PM_SUSPEND
+#define PSMG_ON PM_ON
+#else
+typedef u32 pm_message_t;
+#define PMSG_SUSPEND 3
+#define PMSG_ON 0
+#endif
+#endif
+
+#ifndef PCI_D0
+#define PCI_D0 0
+#define PCI_D3hot 3
+#endif
+
+#ifndef PM_EVENT_SUSPEND
+#define PM_EVENT_SUSPEND 2
+#endif
+
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
+#define pci_choose_state(dev, state) (state)
+#endif
+
 #if defined(CONFIG_SOFTWARE_SUSPEND2) || defined(CONFIG_SUSPEND2)
 #define KTHREAD_RUN(a,b,c) kthread_run(a,b,0,c)
 #else
 #define KTHREAD_RUN(a,b,c) kthread_run(a,b,c)
 #endif
 
+#if !defined(HAVE_NETDEV_PRIV)
+#define netdev_priv(dev)  ((dev)->priv)
+#endif
+
 #ifdef CONFIG_X86_64
 #define LIN2WIN1(func, arg1)			\
 	lin_to_win1(func, (unsigned long)arg1)
@@ -284,11 +331,11 @@ do {									\
 #endif
 
 #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,9)
-#define NW_MODULE_PARM_INT(name, perm) module_param(name, int, perm)
-#define NW_MODULE_PARM_STRING(name, perm) module_param(name, charp, perm)
+#define WRAP_MODULE_PARM_INT(name, perm) module_param(name, int, perm)
+#define WRAP_MODULE_PARM_STRING(name, perm) module_param(name, charp, perm)
 #else
-#define NW_MODULE_PARM_INT(name, perm) MODULE_PARM(name, "i")
-#define NW_MODULE_PARM_STRING(name, perm) MODULE_PARM(name, "s")
+#define WRAP_MODULE_PARM_INT(name, perm) MODULE_PARM(name, "i")
+#define WRAP_MODULE_PARM_STRING(name, perm) MODULE_PARM(name, "s")
 #endif
 
 /* this ugly hack is to handle RH kernels; I don't know any better,
@@ -300,7 +347,7 @@ do {									\
 #define KMALLOC_THRESHOLD 131072
 
 /* TICK is 100ns */
-#define TICKSPERSEC		10000000
+#define TICKSPERSEC		10000000LL
 #define TICKSPERMSEC		10000
 #define SECSPERDAY		86400
 #define SECSPERHOUR		3600
@@ -311,6 +358,22 @@ do {									\
 #define SECS_1601_TO_1970	((369 * 365 + 89) * (u64)SECSPERDAY)
 #define TICKS_1601_TO_1970	(SECS_1601_TO_1970 * TICKSPERSEC)
 
+/* 100ns units to HZ; if sys_time is negative, relative to current
+ * clock, otherwise from year 1601 */
+#define SYSTEM_TIME_TO_HZ(sys_time)					\
+	((((sys_time) <= 0) ? (((u64)HZ * (-(sys_time))) / TICKSPERSEC) : \
+	  (((u64)HZ * ((sys_time) - ticks_1601())) / TICKSPERSEC)))
+
+#define MSEC_TO_HZ(ms) ((ms) * HZ / 1000)
+#define USEC_TO_HZ(ms) ((us) * HZ / 1000000)
+
+extern u64 wrap_ticks_to_boot;
+
+static inline u64 ticks_1601(void)
+{
+	return wrap_ticks_to_boot + (u64)jiffies * TICKSPERSEC / HZ;
+}
+
 typedef void (*WRAP_EXPORT_FUNC)(void);
 
 struct wrap_export {
@@ -333,8 +396,11 @@ struct wrap_export {
 #define WRAP_EXPORT_MAP(s,f)
 #define WRAP_EXPORT(x) x
 
+#define POOL_TAG(A, B, C, D)					\
+	((ULONG)((A) + ((B) << 8) + ((C) << 16) + ((D) << 24)))
+
 struct wrap_alloc {
-	struct list_head list;
+	struct nt_list list;
 	void *ptr;
 };
 
@@ -352,56 +418,84 @@ struct pe_image {
 extern KSPIN_LOCK atomic_lock;
 extern KSPIN_LOCK cancel_lock;
 
-#define DEBUG_IRQL 1
+//#define DEBUG_IRQL 1
 
-#define WRAPPER_TIMER_MAGIC 47697249
-struct wrapper_timer {
-	/* wrapper_timer replaces kdpc field in ktimer; if some
-	 * (nasty) driver passes ktimer->kdpc to other functions,
-	 * having kdpc as first field of wrapper_timer will work as
-	 * expected */
-	struct kdpc *kdpc;
+struct wrap_timer {
+	long repeat;
 	struct nt_list list;
 	struct timer_list timer;
-	struct ktimer *ktimer;
+	struct nt_timer *nt_timer;
 #ifdef DEBUG_TIMER
-	unsigned long wrapper_timer_magic;
+	unsigned long wrap_timer_magic;
 #endif
-	long repeat;
-	int active;
-	KSPIN_LOCK lock;
+};
+
+struct wrap_work_item {
+	struct nt_list list;
+	void *arg1;
+	void *arg2;
+	void *func;
+	BOOLEAN win_func;
 };
 
 typedef struct mdl ndis_buffer;
 
+#define MAX_ALLOCATED_URBS 15
+
 struct phys_dev {
 	int dev_type;
 	struct pci_dev *pci;
-	struct usb_device *usb;
+	struct {
+		struct usb_device *udev;
+		struct usb_interface *intf;
+		int num_alloc_urbs;
+		struct nt_list wrap_urb_list;
+	} usb;
 };
 
+struct wrapper_dev;
+
+#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
+/* until issues with threads hogging cpu are resolved, we don't want
+ * to use shared workqueue, lest the threads take keyboard etc down */
+#define USE_OWN_WORKQUEUE 1
+extern struct workqueue_struct *ndiswrapper_wq;
+#define schedule_work(work_struct) queue_work(ndiswrapper_wq, (work_struct))
+#endif
+
 int ntoskernel_init(void);
 void ntoskernel_exit(void);
+int ntoskernel_init_device(struct wrapper_dev *wd);
+void ntoskernel_exit_device(struct wrapper_dev *wd);
+void *allocate_object(ULONG size, enum common_object_type type,
+		      struct unicode_string *name);
+void  free_object(void *object);
+
 struct driver_object *find_bus_driver(const char *name);
-struct device_object *alloc_pdo(struct driver_object *drv_obj,
-				struct phys_dev *dev);
+struct device_object *alloc_pdo(struct driver_object *drv_obj);
+void free_pdo(struct device_object *drv_obj);
+
+STDCALL void WRITE_PORT_UCHAR(ULONG_PTR port, UCHAR value);
+STDCALL UCHAR READ_PORT_UCHAR(ULONG_PTR port);
 
 STDCALL void *ExAllocatePoolWithTag(enum pool_type pool_type, SIZE_T size,
 				    ULONG tag);
 STDCALL void ExFreePool(void *p);
 STDCALL ULONG MmSizeOfMdl(void *base, ULONG length);
-STDCALL void KeInitializeEvent(struct kevent *kevent,
+STDCALL void *MmMapIoSpace(PHYSICAL_ADDRESS phys_addr, SIZE_T size,
+			   enum memory_caching_type cache);
+STDCALL void MmUnmapIoSpace(void *addr, SIZE_T size);
+STDCALL void KeInitializeEvent(struct nt_event *nt_event,
 			       enum event_type type, BOOLEAN state);
-STDCALL LONG KeSetEvent(struct kevent *kevent, KPRIORITY incr, BOOLEAN wait);
-STDCALL LONG KeResetEvent(struct kevent *kevent);
-STDCALL void KeClearEvent(struct kevent *kevent);
+STDCALL LONG KeSetEvent(struct nt_event *nt_event, KPRIORITY incr,
+			BOOLEAN wait);
+STDCALL LONG KeResetEvent(struct nt_event *nt_event);
+STDCALL void KeClearEvent(struct nt_event *nt_event);
 STDCALL void KeInitializeDpc(struct kdpc *kdpc, void *func, void *ctx);
-BOOLEAN insert_kdpc_work(struct kdpc *kdpc);
-BOOLEAN remove_kdpc_work(struct kdpc *kdpc);
 STDCALL BOOLEAN KeInsertQueueDpc(struct kdpc *kdpc, void *arg1, void *arg2);
 STDCALL BOOLEAN KeRemoveQueueDpc(struct kdpc *kdpc);
-STDCALL NTSTATUS KeWaitForSingleObject(struct kevent *object,
-				       KWAIT_REASON reason,
+STDCALL void KeFlushQueuedDpcs(void);
+STDCALL NTSTATUS KeWaitForSingleObject(void *object, KWAIT_REASON reason,
 				       KPROCESSOR_MODE waitmode,
 				       BOOLEAN alertable,
 				       LARGE_INTEGER *timeout);
@@ -409,6 +503,7 @@ struct mdl *allocate_init_mdl(void *virt, ULONG length);
 void free_mdl(struct mdl *mdl);
 STDCALL struct mdl *IoAllocateMdl(void *virt, ULONG length, BOOLEAN second_buf,
 				  BOOLEAN charge_quota, struct irp *irp);
+STDCALL void MmBuildMdlForNonPagedPool(struct mdl *mdl);
 STDCALL void IoFreeMdl(struct mdl *mdl);
 STDCALL void NdisFreeBuffer(ndis_buffer *buffer);
 _FASTCALL LONG InterlockedDecrement(FASTCALL_DECL_1(LONG volatile *val));
@@ -430,8 +525,11 @@ STDCALL NTSTATUS IoCreateDevice(struct driver_object *driver,
 				DEVICE_TYPE dev_type,
 				ULONG dev_chars, BOOLEAN exclusive,
 				struct device_object **dev_obj);
+STDCALL NTSTATUS IoCreateSymbolicLink(struct unicode_string *link,
+				      struct unicode_string *dev_name);
 STDCALL void IoDeleteDevice(struct device_object *dev);
 STDCALL void IoDetachDevice(struct device_object *topdev);
+STDCALL struct device_object *IoGetAttachedDevice(struct device_object *dev);
 STDCALL NTSTATUS
 IoAllocateDriverObjectExtension(struct driver_object *drv_obj,
 				void *client_id, ULONG extlen, void **ext);
@@ -439,12 +537,43 @@ STDCALL void *IoGetDriverObjectExtension(struct driver_object *drv,
 					 void *client_id);
 STDCALL struct device_object *IoAttachDeviceToDeviceStack
 	(struct device_object *src, struct device_object *dst);
-STDCALL void KeInitializeEvent(struct kevent *kevent, enum event_type type,
+STDCALL void KeInitializeEvent(struct nt_event *nt_event, enum event_type type,
 			       BOOLEAN state);
 void free_custom_ext(struct driver_extension *drv_obj_ext);
-STDCALL NTSTATUS ndis_add_device(struct driver_object *drv_obj,
-				 struct device_object *pdo);
-ULONGLONG ticks_1601(void);
+STDCALL NTSTATUS AddDevice(struct driver_object *drv_obj,
+			   struct device_object *pdo);
+void DeleteDevice(struct device_object *pdo);
+
+driver_dispatch_t IopInvalidDeviceRequest;
+driver_dispatch_t IopPassIrpDown;
+driver_dispatch_t pdoDispatchInternalDeviceControl;
+driver_dispatch_t pdoDispatchDeviceControl;
+driver_dispatch_t pdoDispatchPnp;
+driver_dispatch_t pdoDispatchPower;
+driver_dispatch_t IopPassIrpDownAndWait;
+driver_dispatch_t fdoDispatchPnp;
+
+STDCALL struct irp *IoAllocateIrp(char stack_size, BOOLEAN charge_quota);
+STDCALL void IoFreeIrp(struct irp *irp);
+STDCALL BOOLEAN IoCancelIrp(struct irp *irp);
+_FASTCALL NTSTATUS IofCallDriver
+	(FASTCALL_DECL_2(struct device_object *dev_obj, struct irp *irp));
+STDCALL struct irp *WRAP_EXPORT(IoBuildSynchronousFsdRequest)
+	(ULONG major_func, struct device_object *dev_obj, void *buf,
+	 ULONG length, LARGE_INTEGER *offset, struct nt_event *event,
+	 struct io_status_block *status);
+STDCALL struct irp *WRAP_EXPORT(IoBuildAsynchronousFsdRequest)
+	(ULONG major_func, struct device_object *dev_obj, void *buf,
+	 ULONG length, LARGE_INTEGER *offset,
+	 struct io_status_block *status);
+STDCALL NTSTATUS PoCallDriver(struct device_object *dev_obj, struct irp *irp);
+
+struct nt_thread *wrap_create_thread(struct task_struct *task);
+void wrap_remove_thread(struct nt_thread *thread);
+u64 ticks_1601(void);
+
+int schedule_wrap_work_item(void *func, void *arg1, void *arg2,
+			    BOOLEAN win_func);
 
 STDCALL KIRQL KeGetCurrentIrql(void);
 STDCALL void KeInitializeSpinLock(KSPIN_LOCK *lock);
@@ -452,6 +581,9 @@ STDCALL void KeAcquireSpinLock(KSPIN_LOCK *lock, KIRQL *irql);
 STDCALL void KeReleaseSpinLock(KSPIN_LOCK *lock, KIRQL oldirql);
 STDCALL KIRQL KeAcquireSpinLockRaiseToDpc(KSPIN_LOCK *lock);
 
+STDCALL void IoAcquireCancelSpinLock(KIRQL *irql);
+STDCALL void IoReleaseCancelSpinLock(KIRQL irql);
+
 _FASTCALL KIRQL KfRaiseIrql(FASTCALL_DECL_1(KIRQL newirql));
 _FASTCALL void KfLowerIrql(FASTCALL_DECL_1(KIRQL oldirql));
 _FASTCALL KIRQL KfAcquireSpinLock(FASTCALL_DECL_1(KSPIN_LOCK *lock));
@@ -461,28 +593,43 @@ _FASTCALL void
 IofCompleteRequest(FASTCALL_DECL_2(struct irp *irp, CHAR prio_boost));
 _FASTCALL void
 KefReleaseSpinLockFromDpcLevel(FASTCALL_DECL_1(KSPIN_LOCK *lock));
+STDCALL void RtlCopyMemory(void *dst, const void *src, SIZE_T length);
 STDCALL NTSTATUS RtlUnicodeStringToAnsiString(struct ansi_string *dst,
-					       struct unicode_string *src,
-					       BOOLEAN dup);
+					      const struct unicode_string *src,
+					      BOOLEAN dup);
 STDCALL NTSTATUS RtlAnsiStringToUnicodeString(struct unicode_string *dst,
-					       struct ansi_string *src,
+					       const struct ansi_string *src,
 					       BOOLEAN dup);
-STDCALL void RtlInitAnsiString(struct ansi_string *dst, CHAR *src);
-STDCALL void RtlInitString(struct ansi_string *dst, CHAR *src);
+STDCALL void RtlInitAnsiString(struct ansi_string *dst, const char *src);
+STDCALL void RtlInitString(struct ansi_string *dst, const char *src);
 STDCALL void RtlInitUnicodeString(struct unicode_string *dest,
 				  const wchar_t *src);
 STDCALL void RtlFreeUnicodeString(struct unicode_string *string);
 STDCALL void RtlFreeAnsiString(struct ansi_string *string);
+STDCALL LONG RtlCompareUnicodeString
+	(const struct unicode_string *s1, const struct unicode_string *s2,
+	 BOOLEAN case_insensitive);
+STDCALL void RtlCopyUnicodeString
+	(struct unicode_string *dst, struct unicode_string *src);
+NOREGPARM SIZE_T _win_wcslen(const wchar_t *s);
 
-void *wrap_kmalloc(size_t size, int flags);
+void *wrap_kmalloc(size_t size);
 void wrap_kfree(void *ptr);
-void wrapper_init_timer(struct ktimer *ktimer, void *handle,
-			struct kdpc *kdpc);
-int wrapper_set_timer(struct wrapper_timer *wrapper_timer,
-		      unsigned long expires, unsigned long repeat,
-		      struct kdpc *kdpc);
-void wrapper_cancel_timer(struct wrapper_timer *wrapper_timer,
-			  BOOLEAN *canceled);
+void wrap_init_timer(struct nt_timer *nt_timer, enum timer_type type,
+		     struct wrapper_dev *wd);
+BOOLEAN wrap_set_timer(struct nt_timer *nt_timer, unsigned long expires_hz,
+		       unsigned long repeat_hz, struct kdpc *kdpc);
+
+STDCALL void KeInitializeTimer(struct nt_timer *nt_timer);
+STDCALL void KeInitializeTimerEx(struct nt_timer *nt_timer,
+				 enum timer_type type);
+STDCALL BOOLEAN KeSetTimerEx(struct nt_timer *nt_timer,
+			     LARGE_INTEGER duetime_ticks, LONG period_ms,
+			     struct kdpc *kdpc);
+STDCALL BOOLEAN KeSetTimer(struct nt_timer *nt_timer,
+			   LARGE_INTEGER duetime_ticks, struct kdpc *kdpc);
+STDCALL BOOLEAN KeCancelTimer(struct nt_timer *nt_timer);
+STDCALL void KeInitializeDpc(struct kdpc *kdpc, void *func, void *ctx);
 
 unsigned long lin_to_win1(void *func, unsigned long);
 unsigned long lin_to_win2(void *func, unsigned long, unsigned long);
@@ -496,6 +643,20 @@ unsigned long lin_to_win6(void *func, unsigned long, unsigned long,
 			  unsigned long, unsigned long, unsigned long,
 			  unsigned long);
 
+STDCALL struct nt_thread *KeGetCurrentThread(void);
+STDCALL NTSTATUS
+ObReferenceObjectByHandle(void *handle, ACCESS_MASK desired_access,
+			  void *obj_type, KPROCESSOR_MODE access_mode,
+			  void **object, void *handle_info);
+
+_FASTCALL LONG ObfReferenceObject(FASTCALL_DECL_1(void *object));
+_FASTCALL void ObfDereferenceObject(FASTCALL_DECL_1(void *object));
+STDCALL NTSTATUS ZwClose(void *object);
+#define ObReferenceObject(object)			\
+	ObfReferenceObject(FASTCALL_ARGS_1(object))
+#define ObDereferenceObject(object)			\
+	ObfDereferenceObject(FASTCALL_ARGS_1(object))
+
 #define MSG(level, fmt, ...)				\
 	printk(level "ndiswrapper (%s:%d): " fmt "\n",	\
 	       __FUNCTION__, __LINE__ , ## __VA_ARGS__)
@@ -503,10 +664,17 @@ unsigned long lin_to_win6(void *func, unsigned long, unsigned long,
 #define ERROR(fmt, ...) MSG(KERN_ERR, fmt , ## __VA_ARGS__)
 #define INFO(fmt, ...) MSG(KERN_INFO, fmt , ## __VA_ARGS__)
 
+#define INFOEXIT(stmt) do { INFO("Exit"); stmt; } while(0)
+
 #define UNIMPL() ERROR("--UNIMPLEMENTED--")
 
 void adjust_user_shared_data_addr(char *driver, unsigned long length);
 
+#define IoCompleteRequest(irp, prio)			\
+	IofCompleteRequest(FASTCALL_ARGS_2(irp, prio));
+#define IoCallDriver(dev, irp)				\
+	IofCallDriver(FASTCALL_ARGS_2(dev, irp));
+
 static inline KIRQL current_irql(void)
 {
 	if (in_atomic() || irqs_disabled())
@@ -544,52 +712,77 @@ static inline void lower_irql(KIRQL oldirql)
  * lock seems to be 0 (presumably in Windows value of unlocked
  * spinlock is 0).
  */
-#define KSPIN_LOCK_UNLOCKED 0
+
+/* define CONFIG_DEBUG_SPINLOCK if a Windows driver is suspected of
+ * obtaining a lock while holding the same lock */
+
+//#ifndef CONFIG_DEBUG_SPINLOCK
+//#define CONFIG_DEBUG_SPINLOCK
+//#endif
+
+#undef CONFIG_DEBUG_SPINLOCK
+
+#ifdef CONFIG_DEBUG_SPINLOCK
+#define KSPIN_LOCK_LOCKED ((ULONG_PTR)get_current())
+#else
 #define KSPIN_LOCK_LOCKED 1
+#endif
+
+#define KSPIN_LOCK_UNLOCKED 0
 
 #define kspin_lock_init(lock) *(lock) = KSPIN_LOCK_UNLOCKED
 
 #ifdef CONFIG_SMP
 
-#ifdef __HAVE_ARCH_CMPXCHG
-
-#define kspin_lock(lock)						\
-	while (cmpxchg(lock, KSPIN_LOCK_UNLOCKED, KSPIN_LOCK_LOCKED) != \
+#define raw_kspin_lock(lock)						\
+	while (cmpxchg((lock), KSPIN_LOCK_UNLOCKED, KSPIN_LOCK_LOCKED) != \
 	       KSPIN_LOCK_UNLOCKED)
 
-#else
+#ifdef CONFIG_DEBUG_SPINLOCK
+#define raw_kspin_unlock(lock)						\
+	__asm__ __volatile__("movw $0,%0"				\
+			     :"=m" (*(lock)) : : "memory")
+#else // DEBUG_SPINLOCK
+#define raw_kspin_unlock(lock)						\
+	__asm__ __volatile__("movb $0,%0"				\
+			     :"=m" (*(lock)) : : "memory")
+#endif // DEBUG_SPINLOCK
 
-extern spinlock_t spinlock_kspin_lock;
-#define kspin_lock(lock)				\
-do {							\
-	while (1) {					\
-		spin_lock(&spinlock_kspin_lock);	\
-		if (*(lock) == KSPIN_LOCK_UNLOCKED)	\
-			break;				\
-		spin_unlock(&spinlock_kspin_lock);	\
-	}						\
-	*(lock) = KSPIN_LOCK_LOCKED;			\
-	spin_unlock(&spinlock_kspin_lock);		\
-} while (0)
+#else // SMP
 
-#endif // __HAVE_ARCH_CMPXCHG
+#define raw_kspin_lock(lock) *(lock) = KSPIN_LOCK_LOCKED
+#define raw_kspin_unlock(lock) *(lock) = KSPIN_LOCK_UNLOCKED
 
-#define kspin_unlock(lock) xchg(lock, KSPIN_LOCK_UNLOCKED)
+#endif // SMP
 
-#else
+#ifdef CONFIG_DEBUG_SPINLOCK
 
-#define kspin_lock(lock) *(lock) = KSPIN_LOCK_LOCKED
-#define kspin_unlock(lock) *(lock) = KSPIN_LOCK_UNLOCKED
+#define kspin_lock(lock)						\
+	do {								\
+		if (*(lock) == KSPIN_LOCK_LOCKED)			\
+			ERROR("eeek: process %p already owns lock %p",	\
+			      get_current(), lock);			\
+		else							\
+			raw_kspin_lock(lock);				\
+	} while (0)
+#define kspin_unlock(lock)						\
+	do {								\
+		if (*(lock) != KSPIN_LOCK_LOCKED)			\
+			ERROR("kspin_lock %p not locked!", (lock));	\
+		raw_kspin_unlock(lock);					\
+	} while (0)
 
-#endif // CONFIG_SMP
+#else // DEBUG_SPINLOCK
+
+#define kspin_lock(lock) raw_kspin_lock(lock)
+#define kspin_unlock(lock) raw_kspin_unlock(lock)
+
+#endif // DEBUG_SPINLOCK
 
 /* raise IRQL to given (higher) IRQL if necessary before locking */
 #define kspin_lock_irql(lock, newirql)					\
 ({									\
 	KIRQL _cur_irql_ = current_irql();				\
-	KSPIN_LOCK _val_ = *(lock);					\
-	if (_val_ > KSPIN_LOCK_LOCKED)					\
-		ERROR("illegal spinlock: %p(%lu)", lock, _val_);	\
 	if (_cur_irql_ < DISPATCH_LEVEL && newirql == DISPATCH_LEVEL) {	\
 		local_bh_disable();					\
 		preempt_disable();					\
@@ -602,21 +795,15 @@ do {							\
 #define kspin_unlock_irql(lock, oldirql)				\
 do {									\
 	KIRQL _cur_irql_ = current_irql();				\
-	KSPIN_LOCK _val_ = *(lock);					\
-	if (_val_ > KSPIN_LOCK_LOCKED)					\
-		ERROR("illegal spinlock: %p(%lu)", lock, _val_);	\
 	kspin_unlock(lock);						\
 	if (oldirql < DISPATCH_LEVEL && _cur_irql_ == DISPATCH_LEVEL) {	\
-		preempt_enable();					\
+		preempt_enable_no_resched();				\
 		local_bh_enable();					\
 	}								\
 } while (0)
 
 #define kspin_lock_irqsave(lock, flags)					\
 do {									\
-	KSPIN_LOCK _val_ = *(lock);					\
-	if (_val_ > KSPIN_LOCK_LOCKED)					\
-		ERROR("illegal spinlock: %p(%lu)", lock, _val_);	\
 	local_irq_save(flags);						\
 	preempt_disable();						\
 	kspin_lock(lock);						\
@@ -624,9 +811,6 @@ do {									\
 
 #define kspin_unlock_irqrestore(lock, flags)				\
 do {									\
-	KSPIN_LOCK _val_ = *(lock);					\
-	if (_val_ > KSPIN_LOCK_LOCKED)					\
-		ERROR("illegal spinlock: %p(%lu)", lock, _val_);	\
 	kspin_unlock(lock);						\
 	local_irq_restore(flags);					\
 	preempt_enable();						\
@@ -660,7 +844,7 @@ extern int debug;
 #if defined DEBUG
 #undef DBGTRACE
 #define DBGTRACE(level, fmt, ...) do {					\
-		if (level <= debug)					\
+		if (debug >= level)					\
 			printk(KERN_INFO "%s (%s:%d): " fmt "\n",	\
 			       DRIVER_NAME, __FUNCTION__,		\
 			       __LINE__ , ## __VA_ARGS__);		\
@@ -669,32 +853,32 @@ extern int debug;
 #define DBG_BLOCK()
 #endif
 
-#if defined DEBUG && DEBUG >= 1
+#if defined(DEBUG) && DEBUG >= 1
 #undef DBGTRACE1
 #define DBGTRACE1(fmt, ...) DBGTRACE(1, fmt , ## __VA_ARGS__)
 #endif
 
-#if defined DEBUG && DEBUG >= 2
+#if defined(DEBUG) && DEBUG >= 2
 #undef DBGTRACE2
 #define DBGTRACE2(fmt, ...) DBGTRACE(2, fmt , ## __VA_ARGS__)
 #endif
 
-#if defined DEBUG && DEBUG >= 3
+#if defined(DEBUG) && DEBUG >= 3
 #undef DBGTRACE3
 #define DBGTRACE3(fmt, ...) DBGTRACE(3, fmt , ## __VA_ARGS__)
 #endif
 
-#if defined DEBUG && DEBUG >= 4
+#if defined(DEBUG) && DEBUG >= 4
 #undef DBGTRACE4
 #define DBGTRACE4(fmt, ...) DBGTRACE(4, fmt , ## __VA_ARGS__)
 #endif
 
-#if defined DEBUG && DEBUG >= 5
+#if defined(DEBUG) && DEBUG >= 5
 #undef DBGTRACE5
 #define DBGTRACE5(fmt, ...) DBGTRACE(5, fmt , ## __VA_ARGS__)
 #endif
 
-#if defined DEBUG && DEBUG >= 6
+#if defined(DEBUG) && DEBUG >= 6
 #undef DBGTRACE6
 #define DBGTRACE6(fmt, ...) DBGTRACE(6, fmt , ## __VA_ARGS__)
 #endif
@@ -715,24 +899,70 @@ extern int debug;
 #define TRACEEXIT5(stmt) do { DBGTRACE5("Exit"); stmt; } while(0)
 #define TRACEEXIT6(stmt) do { DBGTRACE6("Exit"); stmt; } while(0)
 
-#if defined(DEBUG) && defined(USB_DEBUG)
-#define USBTRACE(fmt, ...) DBGTRACE1(fmt, ## __VA_ARGS__)
-#define USBTRACEENTER(fmt, ...) TRACEENTER1(fmt, ## __VA_ARGS__)
-#define USBTRACEEXIT(stmt) TRACEEXIT1(stmt)
+//#define USB_DEBUG 1
+//#define EVENT_DEBUG 1
+//#define IO_DEBUG 1
+
+#if defined(USB_DEBUG)
+#define USBTRACE DBGTRACE1
+#define USBENTER TRACEENTER1
+#define USBEXIT TRACEEXIT1
 #else
 #define USBTRACE(fmt, ...)
-#define USBTRACEENTER(fmt, ...)
-#define USBTRACEEXIT(stmt) stmt
+#define USBENTER(fmt, ...)
+#define USBEXIT(stmt) stmt
+#endif
+
+#if defined(EVENT_DEBUG)
+#define EVENTTRACE DBGTRACE1
+#define EVENTENTER TRACEENTER1
+#define EVENTEXIT TRACEEXIT1
+#else
+#define EVENTTRACE(fmt, ...)
+#define EVENTENTER(fmt, ...)
+#define EVENTEXIT(stmt) stmt
+#endif
+
+#if defined(IO_DEBUG)
+#define IOTRACE DBGTRACE1
+#define IOENTER TRACEENTER1
+#define IOEXIT TRACEEXIT1
+#else
+#define IOTRACE(fmt, ...)
+#define IOENTER(fmt, ...)
+#define IOEXIT(stmt) stmt
 #endif
 
 #if defined DEBUG
-#define ASSERT(expr) do {						\
-		if (!(expr)) {						\
-			ERROR("Assertion failed! %s", (#expr));		\
-		}							\
+#define assert(expr) do {						\
+		if (!(expr))						\
+			ERROR("assertion failed: %s", (#expr));		\
 	} while (0)
 #else
-#define ASSERT(expr)
+#define assert(expr) do { } while (0)
 #endif
 
+#if defined(IO_DEBUG)
+#define DUMP_IRP(__irp)							\
+	do {								\
+		struct io_stack_location *_irp_sl;			\
+		_irp_sl = IoGetCurrentIrpStackLocation(__irp);		\
+		IOTRACE("irp: %p, stack size: %d, cl: %d, sl: %p, "	\
+			"dev_obj: %p, mj_fn: %d, minor_fn: %d, "	\
+			"nt_urb: %p, event: %p",			\
+			__irp, __irp->stack_count, (__irp)->current_location, \
+			_irp_sl, _irp_sl->dev_obj, _irp_sl->major_fn,	\
+			_irp_sl->minor_fn, URB_FROM_IRP(__irp),		\
+			(__irp)->user_event);				\
+	} while (0)
+#else
+#define DUMP_IRP(__irp) do { } while (0)
+#endif
+
+#define sleep(nsec)					\
+	do {						\
+		set_current_state(TASK_INTERRUPTIBLE);	\
+		schedule_timeout(nsec * HZ);		\
+	} while (0)
+
 #endif // _NTOSKERNEL_H_
