@@ -217,6 +217,24 @@ static int procfs_read_hw(char *page, char **start, off_t off,
 	if (!res)
 		p += sprintf(p, "rx_antenna=%u\n", antenna);
 
+	p += sprintf(p, "encryption_modes=%s%s%s%s%s%s%s\n",
+		     test_bit(Ndis802_11Encryption1Enabled, &wd->capa.encr) ?
+		     "WEP" : "none",
+
+		     test_bit(Ndis802_11Encryption2Enabled, &wd->capa.encr) ?
+		     "; TKIP with WPA" : "",
+		     test_bit(Ndis802_11AuthModeWPA2, &wd->capa.auth) ?
+		     ", WPA2" : "",
+		     test_bit(Ndis802_11AuthModeWPA2PSK, &wd->capa.auth) ?
+		     ", WPA2PSK" : "",
+
+		     test_bit(Ndis802_11Encryption3Enabled, &wd->capa.encr) ?
+		     "; AES/CCMP with WPA" : "",
+		     test_bit(Ndis802_11AuthModeWPA2, &wd->capa.auth) ?
+		     ", WPA2" : "",
+		     test_bit(Ndis802_11AuthModeWPA2PSK, &wd->capa.auth) ?
+		     ", WPA2PSK" : "");
+
 	if (p - page > count) {
 		WARNING("wrote %lu bytes (limit is %u)",
 			(unsigned long)(p - page), count);
@@ -277,6 +295,26 @@ static int procfs_write_settings(struct file *file, const char *buf,
 		hangcheck_del(wd);
 		wd->hangcheck_interval = i * HZ;
 		hangcheck_add(wd);
+	} else if (!strcmp(setting, "check_capa")) {
+		check_capa(wd);
+		printk(KERN_INFO "%s: encryption modes supported: %s%s%s%s%s%s%s\n",
+		       wd->net_dev->name,
+		       test_bit(Ndis802_11Encryption1Enabled, &wd->capa.encr) ?
+		       "WEP" : "none",
+
+		       test_bit(Ndis802_11Encryption2Enabled, &wd->capa.encr) ?
+		       "; TKIP with WPA" : "",
+		       test_bit(Ndis802_11AuthModeWPA2, &wd->capa.auth) ?
+		       ", WPA2" : "",
+		       test_bit(Ndis802_11AuthModeWPA2PSK, &wd->capa.auth) ?
+		       ", WPA2PSK" : "",
+
+		       test_bit(Ndis802_11Encryption3Enabled, &wd->capa.encr) ?
+		       "; AES/CCMP with WPA" : "",
+		       test_bit(Ndis802_11AuthModeWPA2, &wd->capa.auth) ?
+		       ", WPA2" : "",
+		       test_bit(Ndis802_11AuthModeWPA2PSK, &wd->capa.auth) ?
+		       ", WPA2PSK" : "");
 	} else if (!strcmp(setting, "suspend")) {
 		int i;
 
@@ -287,13 +325,52 @@ static int procfs_write_settings(struct file *file, const char *buf,
 		if (i <= 0 || i > 3)
 			return -EINVAL;
 		if (wd->ndis_device->bustype == NDIS_PCI_BUS)
-			ndiswrapper_suspend_pci(wd->dev.pci, i);
+			ndiswrapper_suspend_pci(wd->dev.pci, PMSG_SUSPEND);
+#if defined(CONFIG_USB) && LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
+		else if (wd->ndis_device->bustype == NDIS_USB_BUS)
+			ndiswrapper_suspend_usb(wd->dev.usb.intf,
+						PMSG_SUSPEND);
+#endif
+		else
+			return -EINVAL;
 	} else if (!strcmp(setting, "resume")) {
 		if (wd->ndis_device->bustype == NDIS_PCI_BUS)
 			ndiswrapper_resume_pci(wd->dev.pci);
+#if defined(CONFIG_USB) && LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
+		else if (wd->ndis_device->bustype == NDIS_USB_BUS)
+			ndiswrapper_resume_usb(wd->dev.usb.intf);
+#endif
+		else
+			return -EINVAL;
 	} else if (!strcmp(setting, "reinit")) {
 		if (ndis_reinit(wd))
 			return -EINVAL;
+#ifdef USB_DEBUG
+	} else if (!strcmp(setting, "irp")) {
+		struct irp *irp;
+		struct io_stack_location *irp_sl;
+		NTSTATUS status;
+		int major_fn, minor_fn, n;
+		struct device_object *dev;
+		if (!p)
+			return -EINVAL;
+		p++;
+		n = sscanf(p, "%d,%d,%x", &major_fn, &minor_fn, (int *)&dev);
+		DBGTRACE1("n = %d, mj = %d, mn = %d, dev = %p", n, major_fn,
+			  minor_fn, dev);
+		if (n != 3)
+			return -EINVAL;
+		irp = IoAllocateIrp(dev->stack_size, FALSE);
+		DBGTRACE1("stack size: %d, irp = %p", dev->stack_size, irp);
+		DBGTRACE1("drv_obj: %p", dev->drv_obj);
+		irp_sl = IoGetNextIrpStackLocation(irp);
+		irp_sl->major_fn = major_fn;
+		irp_sl->minor_fn = minor_fn;
+		irp->io_status.status = STATUS_NOT_SUPPORTED;
+		status = 0;
+		status = IoCallDriver(dev, irp);
+		DBGTRACE1("status = %d", status);
+#endif
 	} else if (!strcmp(setting, "power_profile")) {
 		int i;
 		struct miniport_char *miniport;
@@ -343,6 +420,55 @@ static int procfs_write_settings(struct file *file, const char *buf,
 
 		if (set_encr_mode(wd, i))
 			return -EINVAL;
+	} else if (!strcmp(setting, "stats_enabled")) {
+		int i;
+
+		if (!p)
+			return -EINVAL;
+		p++;
+		i = simple_strtol(p, NULL, 10);
+		if (i > 0)
+			wd->stats_enabled = TRUE;
+		else
+			wd->stats_enabled = FALSE;
+	} else if (!strcmp(setting, "tx_antenna")) {
+		ndis_antenna antenna;
+		unsigned int i;
+		NTSTATUS res;
+
+		if (!p)
+			return -EINVAL;
+		p++;
+		i = simple_strtol(p, NULL, 16);
+		res = miniport_query_info(wd, OID_802_11_NUMBER_OF_ANTENNAS,
+					  &antenna, sizeof(antenna));
+		if (res)
+			return -EINVAL;
+		if (i >= -1 && i < antenna)
+			antenna = i;
+		res = miniport_set_info(wd, OID_802_11_TX_ANTENNA_SELECTED,
+				  &antenna, sizeof(antenna));
+		if (res)
+			return -EINVAL;
+	} else if (!strcmp(setting, "rx_antenna")) {
+		ndis_antenna antenna;
+		unsigned int i;
+		NTSTATUS res;
+
+		if (!p)
+			return -EINVAL;
+		p++;
+		i = simple_strtol(p, NULL, 16);
+		res = miniport_query_info(wd, OID_802_11_NUMBER_OF_ANTENNAS,
+					  &antenna, sizeof(antenna));
+		if (res)
+			return -EINVAL;
+		if (i >= -1 && i < antenna)
+			antenna = i;
+		res = miniport_set_info(wd, OID_802_11_RX_ANTENNA_SELECTED,
+				  &antenna, sizeof(antenna));
+		if (res)
+			return -EINVAL;
 	} else {
 		int res = -1;
 		struct device_setting *dev_setting;
@@ -354,6 +480,7 @@ static int procfs_write_settings(struct file *file, const char *buf,
 		list_for_each_entry(dev_setting, &wd->ndis_device->settings,
 				    list) {
 			struct ndis_config_param *param;
+			struct unicode_string *ustring;
 
 			param = &dev_setting->config_param;
 			if (!stricmp(dev_setting->name, setting)) {
@@ -362,8 +489,9 @@ static int procfs_write_settings(struct file *file, const char *buf,
 				memset(dev_setting->value, 0,
 				       MAX_NDIS_SETTING_VALUE_LEN);
 				memcpy(dev_setting->value, p, strlen(p));
+				ustring = &param->data.ustring;
 				if (param->type == NDIS_CONFIG_PARAM_STRING)
-					RtlFreeUnicodeString(&param->data.ustring);
+					RtlFreeUnicodeString(ustring);
 				param->type = NDIS_CONFIG_PARAM_NONE;
 				res = 0;
 			}
@@ -390,7 +518,7 @@ int ndiswrapper_procfs_add_iface(struct wrapper_dev *wd)
 	wd->procfs_iface = proc_iface;
 
 	if (proc_iface == NULL) {
-		ERROR("%s", "Couldn't create proc directory");
+		ERROR("couldn't create proc directory");
 		return -ENOMEM;
 	}
 	proc_iface->uid = proc_uid;
@@ -399,7 +527,7 @@ int ndiswrapper_procfs_add_iface(struct wrapper_dev *wd)
 	procfs_entry = create_proc_entry("hw", S_IFREG | S_IRUSR | S_IRGRP,
 					 proc_iface);
 	if (procfs_entry == NULL) {
-		ERROR("%s", "Couldn't create proc entry for 'hw'");
+		ERROR("couldn't create proc entry for 'hw'");
 		return -ENOMEM;
 	} else {
 		procfs_entry->uid = proc_uid;
@@ -411,7 +539,7 @@ int ndiswrapper_procfs_add_iface(struct wrapper_dev *wd)
 	procfs_entry = create_proc_entry("stats", S_IFREG | S_IRUSR | S_IRGRP,
 					 proc_iface);
 	if (procfs_entry == NULL) {
-		ERROR("%s", "Couldn't create proc entry for 'stats'");
+		ERROR("couldn't create proc entry for 'stats'");
 		return -ENOMEM;
 	} else {
 		procfs_entry->uid = proc_uid;
@@ -423,7 +551,7 @@ int ndiswrapper_procfs_add_iface(struct wrapper_dev *wd)
 	procfs_entry = create_proc_entry("encr", S_IFREG | S_IRUSR | S_IRGRP,
 					 proc_iface);
 	if (procfs_entry == NULL) {
-		ERROR("%s", "Couldn't create proc entry for 'encr'");
+		ERROR("couldn't create proc entry for 'encr'");
 		return -ENOMEM;
 	} else {
 		procfs_entry->uid = proc_uid;
@@ -436,7 +564,7 @@ int ndiswrapper_procfs_add_iface(struct wrapper_dev *wd)
 					 S_IRUSR | S_IRGRP |
 					 S_IWUSR | S_IWGRP, proc_iface);
 	if (procfs_entry == NULL) {
-		ERROR("%s", "Couldn't create proc entry for 'settings'");
+		ERROR("couldn't create proc entry for 'settings'");
 		return -ENOMEM;
 	} else {
 		procfs_entry->uid = proc_uid;
@@ -482,25 +610,27 @@ static int procfs_read_debug(char *page, char **start, off_t off,
 static int procfs_write_debug(struct file *file, const char *buf,
 			      unsigned long count, void *data)
 {
-	char debug_level[MAX_PROC_STR_LEN];
 	int i;
+	char setting[MAX_PROC_STR_LEN], *p;
 
 	if (count > MAX_PROC_STR_LEN)
 		return -EINVAL;
 
-	memset(debug_level, 0, sizeof(debug_level));
-	if (copy_from_user(debug_level, buf, count))
+	memset(setting, 0, sizeof(setting));
+	if (copy_from_user(setting, buf, count))
 		return -EFAULT;
 
-	i = simple_strtol(debug_level, NULL, 10);
-#if defined(DEBUG) && DEBUG > 0
-	if (i < 0 || i > DEBUG)
-		return -EINVAL;
-#else
-	return -EINVAL;
-#endif
-	debug = i;
+	if ((p = strchr(setting, '\n')))
+		*p = 0;
 
+	if ((p = strchr(setting, '=')))
+		*p = 0;
+
+	i = simple_strtol(setting, NULL, 10);
+	if (i >= 0 && i < 10)
+		debug = i;
+	else
+		return -EINVAL;
 	return count;
 }
 
