@@ -20,9 +20,7 @@
 
 #define IOCTL_INTERNAL_USB_SUBMIT_URB	0x00220003
 #define IOCTL_INTERNAL_USB_RESET_PORT	0x00220007
-
-#define USB_STATUS_SUCCESS		0x00000000
-#define USB_STATUS_ERROR		0x80000000
+#define IOCTL_INTERNAL_USB_GET_PORT_STATUS	0x00220013
 
 #define URB_FUNCTION_SELECT_CONFIGURATION            0x0000
 #define URB_FUNCTION_SELECT_INTERFACE                0x0001
@@ -76,50 +74,95 @@
 #define URB_FUNCTION_SYNC_RESET_PIPE                 0x0030
 #define URB_FUNCTION_SYNC_CLEAR_STALL                0x0031
 
-#define USBD_TRANSFER_DIRECTION_IN	0x00000001
-#define USBD_SHORT_TRANSFER_OK		0x00000002
+#define USBD_PF_CHANGE_MAX_PACKET		0x00000001
+
+#define USBD_TRANSFER_DIRECTION_BIT		0
+#define USBD_SHORT_TRANSFER_OK_BIT		1
+
+#define USBD_TRANSFER_DIRECTION_OUT		0
+#define USBD_TRANSFER_DIRECTION_IN		\
+	(1 << USBD_TRANSFER_DIRECTION_BIT)
+#define USBD_TRANSFER_DIRECTION(flags)		\
+	((flags) & USBD_TRANSFER_DIRECTION_IN)
+
+#define USBD_SHORT_TRANSFER_OK			\
+	(1 << USBD_SHORT_TRANSFER_OK_BIT)
+
+#define USBD_PORT_ENABLED			0x00000001
+#define USBD_PORT_CONNECTED			0x00000002
 
 typedef LONG USBD_STATUS;
 
-union pipe_handle {
-	void *handle;
-	struct {
-		unsigned char endpointAddr;
-		unsigned char pipeType;
-		unsigned char interval;
-		unsigned char fill;
-	} encoded;
-};
+#define USBD_STATUS_SUCCESS			0x0
+#define USBD_STATUS_PENDING			0x40000000
+#define USBD_STATUS_CANCELLED			0x00010000
+
+#define USBD_STATUS_CRC				0xC0000001
+#define USBD_STATUS_BTSTUFF			0xC0000002
+#define USBD_STATUS_DATA_TOGGLE_MISMATCH	0xC0000003
+#define USBD_STATUS_STALL_PID			0xC0000004
+#define USBD_STATUS_DEV_NOT_RESPONDING		0xC0000005
+#define USBD_STATUS_PID_CHECK_FAILURE		0xC0000006
+#define USBD_STATUS_UNEXPECTED_PID	     	0xC0000007
+#define USBD_STATUS_DATA_OVERRUN		0xC0000008
+#define USBD_STATUS_DATA_UNDERRUN		0xC0000009
+#define USBD_STATUS_RESERVED1			0xC000000A
+#define USBD_STATUS_RESERVED2			0xC000000B
+#define USBD_STATUS_BUFFER_OVERRUN		0xC000000C
+#define USBD_STATUS_BUFFER_UNDERRUN		0xC000000D
+#define USBD_STATUS_NOT_ACCESSED		0xC000000F
+#define USBD_STATUS_FIFO			0xC0000010
+#define USBD_STATUS_XACT_ERROR			0xC0000011
+#define USBD_STATUS_BABBLE_DETECTED		0xC0000012
+#define USBD_STATUS_DATA_BUFFER_ERROR		0xC0000013
+#define USBD_STATUS_NOT_SUPPORTED		0xC0000E00
+#define USBD_STATUS_DEVICE_GONE			0xC0007000
+
+#define USBD_STATUS_NO_MEMORY			0x80000100
+#define USBD_STATUS_INVALID_URB_FUNCTION	0x80000200
+#define USBD_STATUS_INVALID_PARAMETER		0x80000300
+#define USBD_STATUS_REQUEST_FAILED		0x80000500
+#define USBD_STATUS_INVALID_PIPE_HANDLE		0x80000600
+#define USBD_STATUS_ERROR_SHORT_TRANSFER	0x80000900
+
+#define USBD_DEFAULT_MAXIMUM_TRANSFER_SIZE	PAGE_SIZE
+
+typedef struct usb_endpoint_descriptor *usbd_pipe_handle;
 
 struct urb_hcd_area {
 	void *reserved8[8];
 };
 
+enum pipe_type {ptControl = USB_ENDPOINT_XFER_CONTROL,
+		ptIsoc = USB_ENDPOINT_XFER_ISOC,
+		ptBulk = USB_ENDPOINT_XFER_BULK,
+		ptInt = USB_ENDPOINT_XFER_INT};
+
 struct usbd_pipe_information {
-	USHORT maxPacketSize;
-	UCHAR endpointAddr;
-	UCHAR interval;
-	enum {ptControl, ptIsochronous, ptBulk, ptIntr} pipeType;
-	union pipe_handle pipeHandle;
-	ULONG maxTransferSize;
-	ULONG pipeFlags;
+	USHORT wMaxPacketSize;
+	UCHAR bEndpointAddress;
+	UCHAR bInterval;
+	enum pipe_type type;
+	usbd_pipe_handle handle;
+	ULONG max_tx_size;
+	ULONG flags;
 };
 
 struct usbd_interface_information {
-	USHORT length;
-	UCHAR intfNum;
-	UCHAR altSet;
-	UCHAR class;
-	UCHAR subClass;
-	UCHAR proto;
-	UCHAR fill;
-	void *intfHandle;
-	ULONG pipeNum;
+	USHORT bLength;
+	UCHAR bInterfaceNumber;
+	UCHAR bAlternateSetting;
+	UCHAR bInterfaceClass;
+	UCHAR bInterfaceSubClass;
+	UCHAR bInterfaceProtocol;
+	UCHAR reserved;
+	void *handle;
+	ULONG bNumEndpoints;
 	struct usbd_pipe_information pipes[1];
 };
 
 struct usbd_interface_list_entry {
-	struct usb_interface_descriptor *intfDesc;
+	struct usb_interface_descriptor *intf_desc;
 	struct usbd_interface_information *intf;
 };
 
@@ -127,59 +170,59 @@ struct nt_urb_header {
 	USHORT length;
 	USHORT function;
 	USBD_STATUS status;
-	void *usbdDevHandle;
-	ULONG usbdFlags;
+	void *usbd_dev_handle;
+	ULONG usbd_flags;
 };
 
-struct select_configuration {
+struct usbd_select_configuration {
 	struct nt_urb_header header;
 	struct usb_config_descriptor *config;
-	void *configHandle;
+	void *handle;
 	struct usbd_interface_information intf;
 };
 
-struct bulk_or_intr_transfer {
+struct usbd_bulk_or_intr_transfer {
 	struct nt_urb_header header;
-	union pipe_handle pipeHandle;
-	ULONG transferFlags;
-	ULONG transferBufLen;
-	void *transferBuf;
-	struct mdl *transferBufMdl;
-	union nt_urb *urbLink;
+	usbd_pipe_handle pipe_handle;
+	ULONG transfer_flags;
+	ULONG transfer_buffer_length;
+	void *transfer_buffer;
+	struct mdl *mdl;
+	union nt_urb *urb_link;
 	struct urb_hcd_area hca;
 };
 
-struct control_descriptor_request {
+struct usbd_control_descriptor_request {
 	struct nt_urb_header header;
 	void *reserved;
 	ULONG reserved0;
-	ULONG transferBufLen;
-	void *transferBuf;
-	struct mdl *transferBufMdl;
-	union nt_urb *urbLink;
+	ULONG transfer_buffer_length;
+	void *transfer_buffer;
+	struct mdl *mdl;
+	union nt_urb *urb_link;
 	struct urb_hcd_area hca;
 	USHORT reserved1;
 	UCHAR index;
-	UCHAR desctype;
-	USHORT langid;
+	UCHAR desc_type;
+	USHORT language_id;
 	USHORT reserved2;
 };
 
-struct pipe_request {
+struct usbd_pipe_request {
 	struct nt_urb_header header;
-	union pipe_handle pipeHandle;
+	usbd_pipe_handle pipe_handle;
 };
 
-struct vendor_or_class_request {
+struct usbd_vendor_or_class_request {
 	struct nt_urb_header header;
 	void *reserved;
-	ULONG transferFlags;
-	ULONG transferBufLen;
-	void *transferBuf;
-	struct mdl *transferBufMdl;
-	union nt_urb *urbLink;
+	ULONG transfer_flags;
+	ULONG transfer_buffer_length;
+	void *transfer_buffer;
+	struct mdl *mdl;
+	union nt_urb *link;
 	struct urb_hcd_area hca;
-	UCHAR reservedBits;
+	UCHAR reserved_bits;
 	UCHAR request;
 	USHORT value;
 	USHORT index;
@@ -192,33 +235,36 @@ struct usbd_iso_packet_desc {
 	USBD_STATUS status;
 };
 
-struct isochronous_transfer {
+struct usbd_isochronous_transfer {
 	struct nt_urb_header header;
-	union pipe_handle pipeHandle;
-	ULONG transferFlags;
-	ULONG transferBufLen;
-	void *transferBuf;
-	struct mdl *transferMDL;
-	union nt_urb *urbLink;
+	usbd_pipe_handle pipe_handle;
+	ULONG transfer_flags;
+	ULONG transfer_buffer_length;
+	void *transfer_buffer;
+	struct mdl *mdl;
+	union nt_urb *urb_link;
 	struct urb_hcd_area hca;
-	ULONG startFrame;
-	ULONG numPackets;
-	ULONG errorCount;
-	struct usbd_iso_packet_desc isoPacket[1];
+	ULONG start_frame;
+	ULONG number_of_packets;
+	ULONG error_count;
+	struct usbd_iso_packet_desc iso_packet[1];
 };
 
 union nt_urb {
 	struct nt_urb_header header;
-	struct select_configuration selConf;
-	struct bulk_or_intr_transfer bulkIntrTrans;
-	struct control_descriptor_request ctrlDescReq;
-	struct vendor_or_class_request venClsReq;
-	struct isochronous_transfer isochTrans;
-	struct pipe_request pipeReq;
+	struct usbd_select_configuration select_conf;
+	struct usbd_bulk_or_intr_transfer bulk_int_transfer;
+	struct usbd_control_descriptor_request control_request;
+	struct usbd_vendor_or_class_request vendor_class_request;
+	struct usbd_isochronous_transfer isochronous;
+	struct usbd_pipe_request pipe_req;
 };
 
-unsigned long usb_submit_nt_urb(struct usb_device *dev, union nt_urb *nt_urb,
-                                struct irp *irp);
-unsigned long usb_reset_port(struct usb_device *dev);
+#define NT_URB_STATUS(nt_urb) ((nt_urb)->header.status)
+
+NTSTATUS wrap_submit_irp(struct device_object *pdo, struct irp *irp);
+NTSTATUS wrap_submit_urb(struct irp *irp);
+void wrap_suspend_urbs(struct wrapper_dev *wd);
+void wrap_resume_urbs(struct wrapper_dev *wd);
 
 #endif /* USB_H */
