@@ -16,11 +16,11 @@
 #ifndef WINNT_TYPES_H
 #define WINNT_TYPES_H
 
-#define TRUE 1
-#define FALSE 0
+#define TRUE				1
+#define FALSE				0
 
-#define PASSIVE_LEVEL 0
-#define DISPATCH_LEVEL 2
+#define PASSIVE_LEVEL			0
+#define DISPATCH_LEVEL			2
 
 #define STATUS_WAIT_0			0
 #define STATUS_SUCCESS                  0
@@ -28,41 +28,78 @@
 #define STATUS_TIMEOUT                  0x00000102
 #define STATUS_PENDING                  0x00000103
 #define STATUS_FAILURE                  0xC0000001
+#define STATUS_NOT_IMPLEMENTED		0xC0000002
 #define STATUS_INVALID_PARAMETER        0xC000000D
+#define STATUS_INVALID_DEVICE_REQUEST	0xC0000010
 #define STATUS_MORE_PROCESSING_REQUIRED 0xC0000016
 #define STATUS_ACCESS_DENIED            0xC0000022
 #define STATUS_BUFFER_TOO_SMALL         0xC0000023
 #define STATUS_RESOURCES                0xC000009A
-#define STATUS_NOT_SUPPORTED            0xC00000BB
-#define STATUS_INVALID_PARAMETER        0xC000000D
+#define STATUS_DELETE_PENDING		0xC0000056
 #define STATUS_INSUFFICIENT_RESOURCES	0xC000009A
+#define STATUS_NOT_SUPPORTED            0xC00000BB
 #define STATUS_INVALID_PARAMETER_2      0xC00000F0
+#define STATUS_NO_MEMORY		0xC0000017
 #define STATUS_CANCELLED                0xC0000120
+#define STATUS_DEVICE_REMOVED		0xC00002B6
 
-#define IS_PENDING                      0x01
-#define CALL_ON_CANCEL                  0x20
-#define CALL_ON_SUCCESS                 0x40
-#define CALL_ON_ERROR                   0x80
+#define SL_PENDING_RETURNED		0x01
+#define CALL_ON_CANCEL			0x20
+#define CALL_ON_SUCCESS			0x40
+#define CALL_ON_ERROR			0x80
 
 #define IRP_MJ_CREATE			0x00
-#define IRP_MJ_CREATE_NAMED_PIPE        0x01
-#define IRP_MJ_CLOSE                    0x02
+#define IRP_MJ_CREATE_NAMED_PIPE	0x01
+#define IRP_MJ_CLOSE			0x02
+#define IRP_MJ_READ			0x03
+#define IRP_MJ_WRITE			0x04
 
-#define IRP_MJ_DEVICE_CONTROL           0x0E
-#define IRP_MJ_INTERNAL_DEVICE_CONTROL  0x0F
-#define IRP_MJ_MAXIMUM_FUNCTION           0x1b
+#define IRP_MJ_DEVICE_CONTROL		0x0E
+#define IRP_MJ_INTERNAL_DEVICE_CONTROL	0x0F
+#define IRP_MJ_POWER			0x16
+#define IRP_MJ_PNP			0x1b
+#define IRP_MJ_MAXIMUM_FUNCTION		0x1b
 
-#define THREAD_WAIT_OBJECTS 3
-#define MAX_WAIT_OBJECTS 64
+#define IRP_MN_WAIT_WAKE		0x00
+#define IRP_MN_POWER_SEQUENCE		0x01
+#define IRP_MN_SET_POWER		0x02
+#define IRP_MN_QUERY_POWER		0x03
 
-#define NOTIFICATION_TIMER 1
+#define IRP_MN_REGINFO			0x08
+#define IRP_MN_REGINFO_EX		0x0b
 
-#define LOW_PRIORITY 		1
-#define LOW_REALTIME_PRIORITY	16
-#define HIGH_PRIORITY		32
-#define MAXIMUM_PRIORITY	32
+#define IRP_MN_START_DEVICE		0x00
+#define IRP_MN_QUERY_REMOVE_DEVICE	0x01
+#define IRP_MN_REMOVE_DEVICE		0x02
+#define IRP_MN_CANCEL_REMOVE_DEVICE	0x03
+#define IRP_MN_STOP_DEVICE		0x04
+#define IRP_MN_QUERY_STOP_DEVICE	0x05
+#define IRP_MN_CANCEL_STOP_DEVICE	0x06
 
-#define PROCESSOR_FEATURE_MAX 64
+#define IRP_BUFFERED_IO			0x00000010
+#define IRP_DEALLOCATE_BUFFER		0x00000020
+
+#define IRP_DEFFER_IO_COMPLETION	0x00000800
+
+#define THREAD_WAIT_OBJECTS		3
+#define MAX_WAIT_OBJECTS		64
+
+#define LOW_PRIORITY			1
+#define LOW_REALTIME_PRIORITY		16
+#define HIGH_PRIORITY			32
+#define MAXIMUM_PRIORITY		32
+
+#define PROCESSOR_FEATURE_MAX		64
+
+#define IO_NO_INCREMENT			0
+
+#define WMIREG_ACTION_REGISTER		1
+#define WMIREG_ACTION_DEREGISTER	2
+#define WMIREG_ACTION_REREGISTER	3
+#define WMIREG_ACTION_UPDATE_GUIDS	4
+
+#define WMIREGISTER			0
+#define WMIUPDATE			1
 
 #ifdef CONFIG_X86_64
 #define STDCALL
@@ -75,9 +112,6 @@
 #define FASTCALL_ARGS_3(arg1,arg2,arg3) arg1, arg2, arg3
 
 #define KI_USER_SHARED_DATA 0xfffff78000000000
-#define SHARED_INTERRUPT_TIME (((char *)&kuser_shared_data + 0x8))
-#define SHARED_SYSTEM_TIME (((char *)&kuser_shared_data + 0x14))
-#define SHARED_TICK_COUNT (((char *)&kuser_shared_data + 0x320))
 
 #else 
 
@@ -116,9 +150,9 @@ typedef u64	ULONGULONG;
 typedef u64	ULONG64;
 
 typedef CHAR CCHAR;
-typedef SHORT wchar_t;
+typedef USHORT wchar_t;
 typedef SHORT CSHORT;
-typedef long long LARGE_INTEGER;
+typedef LONGLONG LARGE_INTEGER;
 
 typedef LONG NTSTATUS;
 
@@ -135,15 +169,18 @@ typedef ULONG_PTR SIZE_T;
 typedef ULONG_PTR KAFFINITY;
 typedef ULONG ACCESS_MASK;
 
+/* non-negative numbers indicate success */
+#define NT_SUCCESS(status)  ((NTSTATUS)(status) >= 0)
+
 struct ansi_string {
-	USHORT len;
-	USHORT buflen;
+	USHORT length;
+	USHORT max_length;
 	char *buf;
 };
 
 struct unicode_string {
-	USHORT len;
-	USHORT buflen;
+	USHORT length;
+	USHORT max_length;
 	wchar_t *buf;
 };
 
@@ -173,6 +210,7 @@ typedef STDCALL void (*DPC)(struct kdpc *kdpc, void *ctx, void *arg1,
 
 struct kdpc {
 	SHORT type;
+	/* number is used to represent if the dpc is queued or not */
 	UCHAR number;
 	UCHAR importance;
 	struct nt_list list;
@@ -248,10 +286,8 @@ struct mdl {
 		(mdl)->next = NULL;					\
 		(mdl)->size = MmSizeOfMdl(baseva, length);		\
 		(mdl)->flags = 0;					\
-		(mdl)->startva = (void *)((ULONG_PTR)baseva &		\
-					  ~(PAGE_SIZE - 1));		\
-		(mdl)->byteoffset = (ULONG)((ULONG_PTR)baseva &		\
-					    (PAGE_SIZE - 1));		\
+		(mdl)->startva = baseva;				\
+		(mdl)->byteoffset = 0;					\
 		(mdl)->bytecount = length;				\
 	}
 
@@ -281,35 +317,51 @@ struct wait_context_block {
 
 struct dispatch_header {
 	UCHAR type;
+	/* 'absolute' field is used by ndiswrapper differently - to
+	 * store type of object this header is associated with */
 	UCHAR absolute;
 	UCHAR size;
 	UCHAR inserted;
 	LONG signal_state;
-	struct nt_list wait_list;
+	struct nt_list wait_blocks;
+};
+
+/* type of object a dh is associated with */
+enum dh_type {
+	DH_NONE, DH_NT_EVENT, DH_NT_TIMER, DH_NT_MUTEX, DH_NT_SEMAPHORE,
+	DH_NT_THREAD,
 };
 
 /* objects that use dispatch_header have it as the first field, so
  * whenever we need to initialize dispatch_header, we can convert that
- * object into a kevent and access dispatch_header */
-struct kevent {
+ * object into a nt_event and access dispatch_header */
+struct nt_event {
 	struct dispatch_header dh;
 };
 
-struct wrapper_timer;
-struct ktimer {
+struct wrap_timer;
+
+#define WRAP_TIMER_MAGIC 47697249
+
+struct nt_timer {
 	struct dispatch_header dh;
-	ULONGLONG due_time;
-	struct nt_list list;
 	/* We can't fit Linux timer in this structure. Instead of
-	 * padding the ktimer structure, we replace *kdpc field with
-	 * *wrapper_timer and allocate memory for it when ktimer is
+	 * padding the nt_timer structure, we replace due_time field
+	 * with *wrap_timer and allocate memory for it when nt_timer is
 	 * initialized */
-	/* struct kdpc *kdpc; */
-	struct wrapper_timer *wrapper_timer;
-	LONG period;
+	union {
+		ULONGLONG due_time;
+		struct wrap_timer *wrap_timer;
+	};
+	struct nt_list list;
+	struct kdpc *kdpc;
+	union {
+		LONG period;
+		LONG wrap_timer_magic;
+	};
 };
 
-struct kmutex {
+struct nt_mutex {
 	struct dispatch_header dh;
 	struct nt_list list;
 	void *owner_thread;
@@ -317,17 +369,40 @@ struct kmutex {
 	BOOLEAN apc_disable;
 };
 
-struct ksemaphore {
+struct nt_semaphore {
 	struct dispatch_header dh;
 	LONG limit;
 };
 
-struct obj_mgr_obj {
+#pragma pack(push,1)
+struct nt_thread {
 	struct dispatch_header dh;
-	struct nt_list list;
-	void *handle;
-	LONG ref_count;
+	/* the rest in Windows is a long structure; since this
+	 * structure is opaque to drivers, we just define what we
+	 * need */
+	int pid;
+	struct task_struct *task;
+	struct nt_list irps;
+	KSPIN_LOCK lock;
+	wait_queue_head_t event_wq;
+	int event_wait_done;
 };
+#pragma pack(pop)
+
+#define set_dh_type(dh, type)		((dh)->absolute = (type))
+#define is_nt_event_dh(dh)		((dh)->absolute == DH_KVENT)
+#define is_nt_timer_dh(dh)		((dh)->absolute == DH_NT_TIMER)
+#define is_mutex_dh(dh)			((dh)->absolute == DH_NT_MUTEX)
+#define is_semaphore_dh(dh)		((dh)->absolute == DH_NT_SEMAPHORE)
+#define is_nt_thread_dh(dh)		((dh)->absolute == DH_NT_THREAD)
+
+#define IO_TYPE_ADAPTER				1
+#define IO_TYPE_CONTROLLER			2
+#define IO_TYPE_DEVICE				3
+#define IO_TYPE_DRIVER				4
+#define IO_TYPE_FILE				5
+#define IO_TYPE_IRP				6
+#define IO_TYPE_DEVICE_OBJECT_EXTENSION		13
 
 struct irp;
 struct dev_obj_ext;
@@ -356,17 +431,11 @@ struct device_object {
 	struct kdpc dpc;
 	ULONG active_threads;
 	void *security_desc;
-	struct kevent lock;
+	struct nt_event lock;
 	USHORT sector_size;
 	USHORT spare1;
 	struct dev_obj_ext *dev_obj_ext;
 	void *reserved;
-
-	/* ndiswrapper-specific data */
-	union {
-		struct usb_device *usb;
-	} device;
-	void *wd;
 };
 
 struct dev_obj_ext {
@@ -384,6 +453,9 @@ struct io_status_block {
 
 struct driver_extension;
 
+typedef NTSTATUS (driver_dispatch_t)(struct device_object *dev_obj,
+				     struct irp *irp) STDCALL;
+
 struct driver_object {
 	CSHORT type;
 	CSHORT size;
@@ -399,7 +471,7 @@ struct driver_object {
 	void *driver_init;
 	void *driver_start_io;
 	void (*driver_unload)(struct driver_object *driver) STDCALL;
-	void *major_func[IRP_MJ_MAXIMUM_FUNCTION + 1];
+	driver_dispatch_t *major_func[IRP_MJ_MAXIMUM_FUNCTION + 1];
 };
 
 struct driver_extension {
@@ -441,8 +513,8 @@ struct file_object {
 	ULONG waiters;
 	ULONG busy;
 	void *last_lock;
-	struct kevent lock;
-	struct kevent event;
+	struct nt_event lock;
+	struct nt_event event;
 	void *completion_context;
 };
 
@@ -452,6 +524,35 @@ struct file_object {
 #define POINTER_ALIGNMENT __attribute__((aligned(8)))
 #endif
 
+enum system_power_state {
+	PowerSystemUnspecified = 0,
+	PowerSystemWorking, PowerSystemSleeping1, PowerSystemSleeping2,
+	PowerSystemSleeping3, PowerSystemHibernate, PowerSystemShutdown,
+	PowerSystemMaximum,
+};
+
+enum device_power_state {
+	PowerDeviceUnspecified = 0,
+	PowerDeviceD0, PowerDeviceD1, PowerDeviceD2, PowerDeviceD3,
+	PowerDeviceMaximum,
+};
+
+union power_state {
+	enum system_power_state system_state;
+	enum device_power_state device_state;
+};
+
+enum power_state_type {
+	SystemPowerState = 0, DevicePowerState,
+};
+
+enum power_action {
+	PowerActionNone = 0,
+	PowerActionReserved, PowerActionSleep, PowerActionHibernate,
+	PowerActionShutdown, PowerActionShutdownReset, PowerActionShutdownOff,
+	PowerActionWarmEject,
+};
+
 #ifndef CONFIG_X86_64
 #pragma pack(push,4)
 #endif
@@ -461,6 +562,7 @@ struct io_stack_location {
 	UCHAR flags;
 	UCHAR control;
 	union {
+		/* NOTE: this union is not complete */
 		struct {
 			void *security_context;
 			ULONG options;
@@ -473,35 +575,54 @@ struct io_stack_location {
 			ULONG POINTER_ALIGNMENT key;
 			LARGE_INTEGER byte_offset;
 		} read;
-		/* FIXME: this structure is not complete */
+		struct {
+			ULONG length;
+			ULONG POINTER_ALIGNMENT key;
+			LARGE_INTEGER byte_offset;
+		} write;
+		struct {
+			ULONG sys_context;
+			enum power_state_type POINTER_ALIGNMENT type;
+			union power_state POINTER_ALIGNMENT state;
+			enum power_action POINTER_ALIGNMENT shutdown_type;
+		} power;
 		struct {
 			ULONG output_buf_len;
-			ULONG input_buf_len; /*align to pointer size*/
-			ULONG code; /*align to pointer size*/
+			ULONG POINTER_ALIGNMENT input_buf_len;
+			ULONG POINTER_ALIGNMENT code;
 			void *type3_input_buf;
 		} ioctl;
+		struct {
+			ULONG_PTR provider_id;
+			void *data_path;
+			ULONG buf_len;
+			void *buf;
+		} wmi;
 		struct {
 			void *arg1;
 			void *arg2;
 			void *arg3;
 			void *arg4;
-		} generic;
+		} others;
 	} params;
 	struct device_object *dev_obj;
 	struct file_object *file_obj;
-	ULONG (*completion_handler)(struct device_object *,
-				    struct irp *, void *) STDCALL;
-	void *handler_arg;
+	NTSTATUS (*completion_routine)(struct device_object *,
+				       struct irp *, void *) STDCALL;
+	void *context;
 };
 #ifndef CONFIG_X86_64
 #pragma pack(pop)
 #endif
 
+#define URB_FROM_IRP(irp)					\
+	(IoGetCurrentIrpStackLocation(irp)->params.others.arg1)
+
 struct kapc {
 	CSHORT type;
 	CSHORT size;
 	ULONG spare0;
-	struct kthread *thread;
+	struct nt_thread *thread;
 	struct nt_list list;
 	void *kernele_routine;
 	void *rundown_routine;
@@ -514,8 +635,28 @@ struct kapc {
 	BOOLEAN inserted;
 };
 
-enum irp_work_type {
-	IRP_WORK_NONE, IRP_WORK_COMPLETE, IRP_WORK_CANCEL,
+#define IRP_NOCACHE			0x00000001
+#define IRP_SYNCHRONOUS_API		0x00000004
+#define IRP_ASSOCIATED_IRP		0x00000008
+
+enum urb_state {
+	URB_INVALID = 1, URB_ALLOCATED, URB_SUBMITTED, URB_CANCELED,
+	URB_COMPLETED, URB_FREE, URB_SUSPEND };
+
+struct wrap_urb {
+	struct nt_list list;
+	enum urb_state state;
+	unsigned int alloc_flags;
+	struct urb *urb;
+	struct irp *irp;
+	unsigned int pipe;
+#ifdef USB_DEBUG
+	unsigned int id;
+#endif
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
+	typeof(((struct urb *)0)->status) urb_status;
+#endif
+
 };
 
 struct irp {
@@ -525,7 +666,7 @@ struct irp {
 	ULONG flags;
 	union {
 		struct irp *master_irp;
-		void *sys_buf;
+		void *system_buffer;
 	} associated_irp;
 
 	struct nt_list threads;
@@ -533,8 +674,8 @@ struct irp {
 	struct io_status_block io_status;
 	KPROCESSOR_MODE requestor_mode;
 	BOOLEAN pending_returned;
-	CHAR stack_size;
-	CHAR stack_pos;
+	CHAR stack_count;
+	CHAR current_location;
 	BOOLEAN cancel;
 	KIRQL cancel_irql;
 
@@ -542,7 +683,7 @@ struct irp {
 	UCHAR alloc_flags;
 
 	struct io_status_block *user_status;
-	struct kevent *user_event;
+	struct nt_event *user_event;
 
 	union {
 		struct {
@@ -561,8 +702,8 @@ struct irp {
 				struct kdevice_queue_entry dev_q_entry;
 				struct {
 					void *driver_context[4];
-				} context;
-			} dev_q;
+				};
+			};
 			void *thread;
 			char *aux_buf;
 			struct {
@@ -571,8 +712,8 @@ struct irp {
 					struct io_stack_location *
 					current_stack_location;
 					ULONG packet_type;
-				} packet;
-			} packet_list;
+				};
+			};
 			struct file_object *file_object;
 		} overlay;
 		struct kapc apc;
@@ -580,60 +721,188 @@ struct irp {
 	} tail;
 
 	/* ndiswrapper extension */
-	enum irp_work_type irp_work_type;
-	struct list_head completed_list;
-	struct list_head cancel_list;
+	struct wrap_urb *wrap_urb;
+	struct nt_list complete_list;
+	struct wrapper_dev *wd;
 };
 
-#define IRP_CUR_STACK_LOC(irp)						\
-	(irp)->tail.overlay.packet_list.packet.current_stack_location
-#define IRP_DRIVER_CONTEXT(irp)					\
-	(irp)->tail.overlay.dev_q.context.driver_context
+#define IoSizeOfIrp(stack_size)						\
+	((USHORT)(sizeof(struct irp) + ((stack_size) *			\
+					sizeof(struct io_stack_location))))
+#define IoGetCurrentIrpStackLocation(irp)		\
+	(irp)->tail.overlay.current_stack_location
+#define IoGetNextIrpStackLocation(irp)		\
+	(IoGetCurrentIrpStackLocation(irp) - 1)
+#define IoGetPreviousIrpStackLocation(irp)	\
+	(IoGetCurrentIrpStackLocation(irp) + 1)
+
+#define IoSetNextIrpStackLocation(IRP)				\
+do {								\
+	(IRP)->current_location--;				\
+	IoGetCurrentIrpStackLocation(IRP)--;			\
+} while (0)
+
+#define IoSkipCurrentIrpStackLocation(IRP) 			\
+do {								\
+	(IRP)->current_location++;				\
+	IoGetCurrentIrpStackLocation(IRP)++;			\
+} while (0)
+
+#define IoCopyCurrentIrpStackLocationToNext(IRP) 			\
+do {									\
+	struct io_stack_location *_irp_sl, *_next;			\
+	_irp_sl = IoGetCurrentIrpStackLocation(IRP);			\
+	_next = IoGetNextIrpStackLocation(IRP);				\
+	RtlCopyMemory(_next, _irp_sl,					\
+		      offsetof(struct io_stack_location,		\
+			       completion_routine));			\
+	_next->control = 0;						\
+} while (0)
+
+#define IoSetCompletionRoutine(IRP, ROUTINE, CONTEXT, SUC, ERR, CANCEL) \
+do {								\
+	struct io_stack_location *__irp_sl;				\
+	__irp_sl = IoGetNextIrpStackLocation(IRP);			\
+	__irp_sl->completion_routine = (ROUTINE);			\
+	__irp_sl->context = (CONTEXT);					\
+	__irp_sl->control = 0;						\
+	if (SUC) { __irp_sl->control |= CALL_ON_SUCCESS; }		\
+	if (ERR) { __irp_sl->control |= CALL_ON_ERROR; }		\
+	if (CANCEL) { __irp_sl->control |= CALL_ON_CANCEL; }		\
+} while (0)
+
+#define IoMarkIrpPending(irp)						\
+	(IoGetCurrentIrpStackLocation((irp))->control |= SL_PENDING_RETURNED)
+
+#define IRP_SL(irp, i) (((struct io_stack_location *)((irp) + 1)) + (i))
+#define IRP_DRIVER_CONTEXT(irp) (irp)->tail.overlay.driver_context
+#define IoIrpThread(irp) ((irp)->tail.overlay.thread)
+
+struct guid {
+	unsigned long data1;
+	unsigned short data2;
+	unsigned short data3;
+	unsigned char data4[8];
+};
+
+struct wmi_guid_reg_info {
+	struct guid *guid;
+	ULONG instance_count;
+	ULONG flags;
+};
+
+struct wmilib_context {
+	ULONG guid_count;
+	struct wmi_guid_reg_info *guid_list;
+	void *query_wmi_reg_info;
+	void *query_wmi_data_block;
+	void *set_wmi_data_block;
+	void *set_wmi_data_item;
+	void *execute_wmi_method;
+	void *wmi_function_control;
+};
+
+enum key_value_information_class {
+	KeyValueBasicInformation, KeyValueFullInformation,
+	KeyValuePartialInformation, KeyValueFullInformationAlign64,
+	KeyValuePartialInformationAlign64
+};
+
+struct object_attr {
+	ULONG length;
+	void *root_dir;
+	struct unicode_string *name;
+	ULONG attr;
+	union {
+		void *security_descriptor;
+		void *file;
+	};
+	void *security_qos;
+};
+
+struct file_name_info {
+	ULONG length;
+	wchar_t *name;
+};
+
+enum file_info_class {
+	FileAlignmentInformation = 17,
+	FileAttributeTagInformation = 35,
+	FileBasicInformation = 4,
+	FileNameInformation = 9,
+	FileNetworkOpenInformation = 34,
+	FilePositionInformation = 14,
+	FileStandardInformation = 5,
+};
 
 enum nt_obj_type {
 	NT_OBJ_EVENT = 10, NT_OBJ_MUTEX, NT_OBJ_THREAD, NT_OBJ_TIMER,
 	NT_OBJ_SEMAPHORE,
 };
 
-struct common_body_header {
-	CSHORT type;
-	CSHORT size;
+enum common_object_type {
+	OBJECT_TYPE_NONE, OBJECT_TYPE_DEVICE, OBJECT_TYPE_DRIVER,
+	OBJECT_TYPE_NT_THREAD, OBJECT_TYPE_FILE, OBJECT_TYPE_CALLBACK,
 };
 
-struct object_header {
-	struct unicode_string name;
+struct common_object_header {
 	struct nt_list list;
-	LONG ref_count;
-	LONG handle_count;
+	enum common_object_type type;
+	int size;
+	struct unicode_string *name;
+	unsigned int ref_count;
 	BOOLEAN close_in_process;
 	BOOLEAN permanent;
-	BOOLEAN inherit;
-	void *parent;
-	void *object_type;
-	void *security_desc;
-	CSHORT type;
-	CSHORT size;
 };
 
+#define OBJECT_TO_HEADER(object)					\
+	(struct common_object_header *)((void *)(object) -		\
+					sizeof(struct common_object_header))
+#define OBJECT_SIZE(size)				\
+	((size) + sizeof(struct common_object_header))
+#define HEADER_TO_OBJECT(hdr)					\
+	((void *)(hdr) + sizeof(struct common_object_header))
+#define HANDLE_TO_OBJECT(handle) HEADER_TO_OBJECT(handle)
+#define HANDLE_TO_HEADER(handle) (handle)
+
 enum work_queue_type {
 	CriticalWorkQueue, DelayedWorkQueue, HyperCriticalWorkQueue,
 	MaximumWorkQueue
 };
 
-enum wait_type {
-	WaitAll, WaitAny
+struct io_workitem {
+	enum work_queue_type type;
+	struct device_object *dev_obj;
+	void (*worker_routine)(struct device_object *dev_obj,
+			       void *context) STDCALL;
+	void *context;
+};
+
+struct io_workitem_entry {
+	struct nt_list list;
+	struct io_workitem *io_workitem;
 };
 
 struct wait_block {
-	struct nt_list list_entry;
-	void *thread;
+	struct nt_list list;
+	struct nt_thread *thread;
 	void *object;
 	struct wait_block *next;
 	USHORT wait_key;
 	USHORT wait_type;
 };
 
-enum event_type {NotificationEvent, SynchronizationEvent};
+enum event_type {
+	NotificationEvent, SynchronizationEvent
+};
+
+enum timer_type {
+	NotificationTimer, SynchronizationTimer
+};
+
+enum wait_type {
+	WaitAll, WaitAny
+};
 
 enum mm_page_priority {
 	LowPagePriority, NormalPagePriority = 16, HighPagePriority = 32
@@ -823,6 +1092,45 @@ struct kuser_shared_data {
 	} tick;
 };
 
+#define REG_NONE			(0)
+#define REG_SZ				(1)
+#define REG_EXPAND_SZ			(2)
+#define REG_BINARY			(3)
+#define REG_DWORD			(4)
+
+#define RTL_REGISTRY_ABSOLUTE		0
+#define RTL_REGISTRY_SERVICES		1
+#define RTL_REGISTRY_CONTROL		2
+#define RTL_REGISTRY_WINDOWS_NT		3
+#define RTL_REGISTRY_DEVICEMAP		4
+#define RTL_REGISTRY_USER		5
+#define RTL_REGISTRY_MAXIMUM		6
+#define RTL_REGISTRY_HANDLE		0x40000000
+#define RTL_REGISTRY_OPTIONAL		0x80000000
+
+#define RTL_QUERY_REGISTRY_SUBKEY	0x00000001
+#define RTL_QUERY_REGISTRY_TOPKEY	0x00000002
+#define RTL_QUERY_REGISTRY_REQUIRED	0x00000004
+#define RTL_QUERY_REGISTRY_NOVALUE	0x00000008
+#define RTL_QUERY_REGISTRY_NOEXPAND	0x00000010
+#define RTL_QUERY_REGISTRY_DIRECT	0x00000020
+#define RTL_QUERY_REGISTRY_DELETE	0x00000040
+
+typedef NTSTATUS (*PRTL_QUERY_REGISTRY_ROUTINE)(wchar_t *name, ULONG type,
+						void *data, ULONG length,
+						void *context,
+						void *entry) STDCALL;
+
+struct rtl_query_registry_table {
+	PRTL_QUERY_REGISTRY_ROUTINE query_func;
+	ULONG flags;
+	wchar_t *name;
+	void *context;
+	ULONG def_type;
+	void *def_data;
+	ULONG def_length;
+};
+
 /* some of the functions below are slightly different from DDK's
  * implementation; e.g., Insert functions return appropriate
  * pointer */
@@ -839,7 +1147,7 @@ static inline void InitializeListHead(struct nt_list *head)
 
 static inline BOOLEAN IsListEmpty(struct nt_list *head)
 {
-	if (head->next == head)
+	if (head == head->next)
 		return TRUE;
 	else
 		return FALSE;
@@ -847,83 +1155,84 @@ static inline BOOLEAN IsListEmpty(struct nt_list *head)
 
 static inline void RemoveEntryList(struct nt_list *entry)
 {
-	struct nt_list *prev, *next;
-
-	next = entry->next;
-	prev = entry->prev;
-	prev->next = next;
-	next->prev = prev;
+	entry->prev->next = entry->next;
+	entry->next->prev = entry->prev;
 }
 
 static inline struct nt_list *RemoveHeadList(struct nt_list *head)
 {
-	struct nt_list *next, *entry;
+	struct nt_list *entry;
 
-	if (IsListEmpty(head))
+	entry = head->next;
+	if (entry == head)
 		return NULL;
 	else {
-		entry = head->next;
-		next = entry->next;
-		head->next = next;
-		next->prev = head;
+		RemoveEntryList(entry);
 		return entry;
 	}
 }
 
 static inline struct nt_list *RemoveTailList(struct nt_list *head)
 {
-	struct nt_list *prev, *entry;
+	struct nt_list *entry;
 
-	if (IsListEmpty(head))
+	entry = head->prev;
+	if (entry == head)
 		return NULL;
 	else {
-		entry = head->prev;
-		prev = entry->prev;
-		head->prev = prev;
-		prev->next = head;
+		RemoveEntryList(entry);
 		return entry;
 	}
 }
 
+static inline void InsertListEntry(struct nt_list *entry, struct nt_list *prev,
+				   struct nt_list *next)
+{
+	next->prev = entry;
+	entry->next = next;
+	entry->prev = prev;
+	prev->next = entry;
+}
+
 static inline struct nt_list *InsertHeadList(struct nt_list *head,
 					     struct nt_list *entry)
 {
-	struct nt_list *next, *first;
+	struct nt_list *ret;
 
 	if (IsListEmpty(head))
-		first = NULL;
+		ret = NULL;
 	else
-		first = head->next;
+		ret = head->next;
 
-	next = head->next;
-	entry->next = next;
-	entry->prev = head;
-	next->prev = entry;
-	head->next = entry;
-	return first;
+	InsertListEntry(entry, head, head->next);
+	return ret;
 }
 
 static inline struct nt_list *InsertTailList(struct nt_list *head,
 					     struct nt_list *entry)
 {
-	struct nt_list *prev, *last;
+	struct nt_list *ret;
 
 	if (IsListEmpty(head))
-		last = NULL;
+		ret = NULL;
 	else
-		last = head->prev;
+		ret = head->prev;
 
-	prev = head->prev;
-	entry->next = head;
-	entry->prev = prev;
-	prev->next = entry;
-	head->prev = entry;
-	return last;
+	InsertListEntry(entry, head->prev, head);
+	return ret;
 }
 
 #define nt_list_for_each(pos, head)					\
-	for (pos = (head)->next; prefetch(pos->next), pos != (head);	\
-	     pos = pos->next)
+	for (pos = (head)->next; pos != (head); pos = pos->next)
+
+#define nt_list_for_each_entry(pos, head, member)			\
+	for (pos = container_of((head)->next, typeof(*pos), member);	\
+	     &pos->member != (head);					\
+	     pos = container_of(pos->member.next, typeof(*pos), member))
+
+#define nt_list_for_each_safe(pos, n, head)		       \
+	for (pos = (head)->next, n = pos->next; pos != (head); \
+	     pos = n, n = pos->next)
 
 static inline struct nt_slist *
 PushEntryList(union nt_slist_head *head, struct nt_slist *entry)
@@ -1041,4 +1350,15 @@ static inline struct nt_slist *PopEntryList(union nt_slist_head *head)
 #define FILE_AUTOGENERATED_DEVICE_NAME	0x00000080
 #define FILE_DEVICE_SECURE_OPEN		0x00000100
 
+#define FILE_READ_DATA			0x0001
+#define FILE_WRITE_DATA			0x0002
+
+#define FILE_SUPERSEDED			0x00000000
+#define FILE_OPENED			0x00000001
+#define FILE_CREATED			0x00000002
+#define FILE_OVERWRITTEN		0x00000003
+#define FILE_EXISTS			0x00000004
+#define FILE_DOES_NOT_EXIST		0x00000005
+
+
 #endif /* WINNT_TYPES_H */
